import typing as tp
from enum import Enum

from functools import partial
from functools import lru_cache
from itertools import chain
from itertools import repeat

from hypothesis import strategies as st
from hypothesis.extra import numpy as hypo_np
from hypothesis import settings as hypo_settings
from hypothesis import HealthCheck


import numpy as np

from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_BOOL
# from static_frame.core.util import DTYPE_INEXACT_KINDS

from static_frame import TypeBlocks

from static_frame import Index
from static_frame import IndexGO
from static_frame import IndexYear
from static_frame import IndexYearGO
from static_frame import IndexYearMonth
from static_frame import IndexYearMonthGO
from static_frame import IndexDate
from static_frame import IndexDateGO
from static_frame import IndexHour
from static_frame import IndexHourGO
from static_frame import IndexMinute
from static_frame import IndexMinuteGO
from static_frame import IndexMillisecond
from static_frame import IndexMillisecondGO
from static_frame import IndexMicrosecond
from static_frame import IndexMicrosecondGO
from static_frame import IndexNanosecond
from static_frame import IndexNanosecondGO
from static_frame import IndexHierarchy
from static_frame import IndexHierarchyGO
from static_frame import Series
from static_frame import Frame
from static_frame import FrameGO


MAX_ROWS = 8
MAX_COLUMNS = 10


hypo_settings.register_profile("sf",
        suppress_health_check=[HealthCheck.too_slow],
        deadline=None,
        )
hypo_settings.load_profile("sf")

#-------------------------------------------------------------------------------
# spacings

def subset_contiguous_sum(target: int) -> tp.Tuple[tp.Tuple[int, ...], ...]:
    '''
    Return an iterabel of integers that sum to the target. This does not find all combinations or permutations, just all combintation of the range from 1 to the number (inclusive).
    '''
    # based on https://stackoverflow.com/questions/4632322/finding-all-possible-combinations-of-numbers-to-reach-a-given-sum

    if target == 0:
        return ()

    if not 0 < target <= 32:
        # over sizes of 60 or so performance is noticieable
        raise RuntimeError(f'target is too large: {target}')

    @lru_cache()
    def subset_sum(
            numbers: tp.Sequence[int], partial: tp.Tuple[int, ...] = (), partial_sum: int = 0,
        ) -> tp.Iterator[tp.Tuple[int, ...]]:
        if partial_sum == target:
            yield partial
        if partial_sum > target:
            return
        for i, n in enumerate(numbers, start=1):
            # get pairs of index (starting at 1) and n (the value at each position)
            yield from subset_sum(numbers[i:], partial + (n,), partial_sum + n)

    return tuple(subset_sum(range(1, target+1)))


def get_spacing(size: int = MAX_COLUMNS) -> st.SearchStrategy:
    # generate permutations of the orderings of the integers
    return st.one_of((st.permutations(c) for c in subset_contiguous_sum(size)))


#-------------------------------------------------------------------------------
# values

# 55203 is just before "high surrogates", and avoids this exception
# UnicodeDecodeError: 'utf-32-le' codec can't decode bytes in position 0-3: code point in surrogate code point range(0xd800, 0xe000)
ST_CODEPOINT_LIMIT = dict(min_codepoint=1, max_codepoint=55203)

ST_TYPES_COMMON: tp.Tuple[tp.Callable[..., st.SearchStrategy], ...] = (
        st.integers,
        # st.decimals,
        st.fractions,
        st.dates,
        st.datetimes,
        partial(st.characters, **ST_CODEPOINT_LIMIT),
        partial(st.text, st.characters(**ST_CODEPOINT_LIMIT))  # type: ignore
        )

ST_TYPES_FLOAT_NAN: tp.Tuple[st.SearchStrategy, ...] = (
        st.floats,
        st.complex_numbers,
        )

filter_nan = lambda x: not np.isnan(x)

ST_TYPES_FLOAT_NO_NAN: tp.Tuple[tp.Callable[[], st.SearchStrategy], ...] = (
        lambda: st.floats().filter(filter_nan),
        lambda: st.complex_numbers().filter(filter_nan)
        )

ST_TYPES_UNARY_BINARY = (st.booleans, st.none)

# common collections

ST_TYPES_FOR_UNIQUE = ST_TYPES_FLOAT_NO_NAN + ST_TYPES_COMMON
ST_TYPES_FOR_UNIQUE_MIXED = ST_TYPES_FLOAT_NO_NAN + ST_TYPES_COMMON + ST_TYPES_UNARY_BINARY
ST_VALUE = ST_TYPES_FLOAT_NAN + ST_TYPES_COMMON + ST_TYPES_UNARY_BINARY

def get_value() -> st.SearchStrategy:
    '''
    Any plausible value.
    '''
    return st.one_of(strat() for strat in ST_VALUE)

def get_label() -> st.SearchStrategy:
    '''
    A hashable suitable for use in an Index. While NaNs are supported as labels in Index objects, the unique constraint used below does not enforce uniqueness for NaNs, and thus we must filter out NaNs in advance.
    '''
    return st.one_of((strat() for strat in ST_TYPES_FOR_UNIQUE_MIXED))


def get_labels(
        min_size: int = 0,
        max_size: int = MAX_ROWS) -> st.SearchStrategy:
    '''
    Labels are suitable for creating non-date Indices (though they might include dates); these labels might force an object array result.
    '''
    def gen() -> tp.Iterator[st.SearchStrategy]:

        yield st.lists(get_label(),
                min_size=min_size,
                max_size=max_size,
                unique=True)

        for strat in ST_TYPES_FOR_UNIQUE:
            yield st.lists(
                    strat(),
                    min_size=min_size,
                    max_size=max_size,
                    unique=True)

    return st.one_of(gen())


#-------------------------------------------------------------------------------
# dtypes


class DTGroup(Enum):
    # NOTE: we branch on these enums in get_array_from_dtype_group to add object dtypes when appropriate

    OBJECT = (partial(st.just, DTYPE_OBJECT),)
    ALL = (hypo_np.scalar_dtypes,)

    NUMERIC = (
            hypo_np.floating_dtypes,
            hypo_np.integer_dtypes,
            hypo_np.complex_number_dtypes
            )

    BOOL = (partial(st.just, DTYPE_BOOL),)
    STRING = (hypo_np.unicode_string_dtypes,)

    YEAR = (partial(hypo_np.datetime64_dtypes, min_period='Y', max_period='Y'),)
    YEAR_MONTH = (partial(hypo_np.datetime64_dtypes, min_period='M', max_period='M'),)
    DATE = (partial(hypo_np.datetime64_dtypes, min_period='D', max_period='D'),)
    HOUR = (partial(hypo_np.datetime64_dtypes, min_period='h', max_period='h'),)
    MINUTE = (partial(hypo_np.datetime64_dtypes, min_period='m', max_period='m'),)
    SECOND = (partial(hypo_np.datetime64_dtypes, min_period='s', max_period='s'),)
    MILLISECOND = (partial(hypo_np.datetime64_dtypes, min_period='ms', max_period='ms'),)
    MICROSECOND = (partial(hypo_np.datetime64_dtypes, min_period='us', max_period='us'),)
    NANOSECOND = (partial(hypo_np.datetime64_dtypes, min_period='ns', max_period='ns'),)

    # derived
    NUMERIC_REAL = (
            hypo_np.floating_dtypes,
            hypo_np.integer_dtypes,
            )
    DATETIME = tuple(chain(
            YEAR,
            YEAR_MONTH,
            DATE,
            HOUR,
            SECOND,
            MILLISECOND,
            MICROSECOND,
            NANOSECOND,
            ))

    BASIC = NUMERIC + BOOL + STRING

    # NOTE: duplicate non-datetime to produce more balanced distribution
    CORE = tuple(chain(
            # OBJECT, # object has to be handled with get_array_from_dtype_group
            NUMERIC, NUMERIC, NUMERIC,
            BOOL, BOOL, BOOL,
            STRING, STRING, STRING,
            DATETIME,
            ))

    ALL_NO_OBJECT = tuple(chain(
            ALL, ALL, ALL,
            NUMERIC, NUMERIC, NUMERIC,
            BOOL, BOOL, BOOL,
            STRING, STRING, STRING,
            DATETIME,
            ))

def get_dtype(dtype_group: DTGroup = DTGroup.ALL) -> st.SearchStrategy:

    def st_dts() -> tp.Iterator[st.SearchStrategy]:
        for st_dt in dtype_group.value:
            yield st_dt()

    return st.one_of(st_dts())

def get_dtypes(
        min_size: int = 0,
        max_size: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL,
        ) -> st.SearchStrategy:
    return st.lists(get_dtype(dtype_group), min_size=min_size)

def get_dtype_pairs(
        dtype_group: DTGroup = DTGroup.ALL,
        ) -> st.SearchStrategy:
    return st.tuples(get_dtype(dtype_group), get_dtype(dtype_group))

#-------------------------------------------------------------------------------
# shape generation

def get_shape_1d(min_size: int = 0, max_size: int = MAX_ROWS) -> st.SearchStrategy:
    return st.tuples(st.integers(min_value=min_size, max_value=max_size))

def get_shape_2d(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        ) -> st.SearchStrategy:
    return st.tuples(
            st.integers(min_value=min_rows, max_value=max_rows),
            st.integers(min_value=min_columns, max_value=max_columns)
            )

def get_shape_1d2d(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS) -> st.SearchStrategy:

    return st.one_of(
            get_shape_2d(
                    min_rows=min_rows,
                    max_rows=max_rows,
                    min_columns=min_columns,
                    max_columns=max_columns),
            get_shape_1d(
                    min_size=min_rows,
                    max_size=max_rows)
            )

#-------------------------------------------------------------------------------
# array generation

def get_array_object(
        shape: tp.Tuple[int, ...] = (MAX_ROWS, MAX_COLUMNS),
        unique: bool = True) -> st.SearchStrategy:
    if unique:
        # if unique, cannot use fill
        return hypo_np.arrays(
                shape=shape,
                dtype=get_dtype(DTGroup.OBJECT),
                elements=get_value(),
                fill=st.nothing(),
                unique=unique
                )
    return hypo_np.arrays(
            shape=shape,
            dtype=get_dtype(DTGroup.OBJECT),
            elements=get_value(),
            fill=st.none(),
            unique=unique
            )


def get_array_from_dtype_group(
        dtype_group: DTGroup,
        shape: tp.Tuple[int, ...] = (MAX_ROWS, MAX_COLUMNS),
        unique: bool = True,
        ) -> st.SearchStrategy:
    '''
    Given a dtype group and shape, get array. Handles manually creating and filling object arrays when dtype group is object or ALL.
    '''

    # TODO: can remove floating-point NaNs when necessary with .map call with this function on array generators; can apply based on DTYPE group

    # def fill_na(array: np.ndarray) -> np.ndarray:
    #     if array.dtype.kind in DTYPE_INEXACT_KINDS:
    #         is_nan = np.isnan(array)
    #         if is_nan.any():
    #             fill = np.empty(array.shape, dtype=array.dtype)
    #             array[is_nan] = fill[is_nan]
    #             return array
    #     return array

    array_object = get_array_object(
            shape=shape,
            unique=unique
            )
    array_non_object = hypo_np.arrays(
            get_dtype(dtype_group),
            shape,
            unique=unique
            )

    if dtype_group is DTGroup.OBJECT:
        return array_object
    if dtype_group is DTGroup.ALL or dtype_group is DTGroup.CORE:
        return st.one_of(array_non_object, array_non_object, array_object)
    return array_non_object


def get_array_1d(
        min_size: int = 0,
        max_size: int = MAX_ROWS,
        unique: bool = False,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:

    shape = get_shape_1d(min_size=min_size, max_size=max_size)
    return get_array_from_dtype_group(
            dtype_group=dtype_group,
            shape=shape,
            unique=unique
            )


def get_array_2d(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        unique: bool = False,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:

    shape = get_shape_2d(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns
            )

    return get_array_from_dtype_group(
            dtype_group=dtype_group,
            shape=shape,
            unique=unique
            )


def get_array_1d2d(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:
    '''
    For convenience in building blocks, treat row constraints as 1d size constraints.
    '''
    array_2d = get_array_2d(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns,
            dtype_group=dtype_group
            )

    if 1 in range(min_columns, max_columns + 1):
        # if min/max columns are given, and column of 1 is not supported, it is incorrect to give back a 1D array (in the context of the usage of this in blocks)
        return st.one_of(
                array_2d,
                get_array_1d(
                        min_size=min_rows,
                        max_size=max_rows,
                        dtype_group=dtype_group
                        )
                )
    return array_2d

#-------------------------------------------------------------------------------
# aligend arrays for concatenation and type blocks

def get_arrays_2d_aligned_columns(
        min_size: int = 1,
        max_size: int = 10) -> st.SearchStrategy:

    return st.integers(min_value=1, max_value=MAX_COLUMNS).flatmap(
        lambda columns: st.lists(
            get_array_2d(
                min_columns=columns,
                max_columns=columns
                ),
            min_size=min_size,
            max_size=max_size
            )
    )

def get_arrays_2d_aligned_rows(
        min_size: int = 1,
        max_size: int = 10) -> st.SearchStrategy:

    return st.integers(min_value=1, max_value=MAX_ROWS).flatmap(
        lambda rows: st.lists(
            get_array_2d(
                min_rows=rows,
                max_rows=rows,
                ),
            min_size=min_size,
            max_size=max_size
            )
    )

def get_arrays_2d_aligned(
        min_size: int = 1,
        max_size: int = 10) -> st.SearchStrategy:

    return get_shape_2d().flatmap(
        lambda shape: st.lists(
            get_array_2d(
                min_rows=shape[0],
                max_rows=shape[0],
                min_columns=shape[1],
                max_columns=shape[1],
                ),
            min_size=min_size,
            max_size=max_size
            )
    )

def get_blocks(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:
    '''
    Args:
        min_columns: number of resultant columns in combination of all arrays.
    '''

    def constructor(rows_column_widths: tp.Tuple[int, tp.Iterator[int]]) -> st.SearchStrategy:
        rows, column_widths = rows_column_widths

        def array_gen() -> tp.Iterator[st.SearchStrategy]:
            for width in column_widths:
                yield get_array_1d2d(
                    min_rows=rows,
                    max_rows=rows,
                    min_columns=width,
                    max_columns=width,
                    dtype_group=dtype_group
                    )

        return st.tuples(*array_gen())

    def get_column_widths(shape: tp.Tuple[int, int]) -> st.SearchStrategy:
        rows, columns = shape
        return st.tuples(st.just(rows), get_spacing(columns)).flatmap(constructor)

    return get_shape_2d(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns).flatmap(get_column_widths)


def get_type_blocks(
        min_rows: int = 0,
        max_rows: int = MAX_ROWS,
        min_columns: int = 0,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:
    return st.builds(TypeBlocks.from_blocks,
            get_blocks(min_rows=min_rows,
                    max_rows=max_rows,
                    min_columns=min_columns,
                    max_columns=max_columns,
                    dtype_group=dtype_group)
            )


get_type_blocks_numeric: tp.Callable[..., st.SearchStrategy] = partial(get_type_blocks, dtype_group=DTGroup.NUMERIC)
get_type_blocks_numeric.__name__ = 'get_type_blocks_numeric'


def get_type_blocks_aligned_array(
        min_rows: int = 0,
        max_rows: int = MAX_ROWS,
        min_columns: int = 0,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:
    '''
    Return TypeBlocks instance, as well as an array aligned by row size.
    '''
    def constructor(shape: tp.Tuple[int, int]) -> st.SearchStrategy:
        rows, columns = shape
        return st.tuples(
                get_type_blocks(
                        min_rows=rows,
                        max_rows=rows,
                        min_columns=columns,
                        max_columns=columns,
                        dtype_group=dtype_group
                        ),
                get_array_1d2d( # let columns be variable3
                        min_rows=rows,
                        max_rows=rows,
                        dtype_group=dtype_group
                        )
                )

    return get_shape_2d(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns,
            ).flatmap(constructor)

def get_type_blocks_aligned_type_blocks(
        min_size: int = 0,
        max_size: int = MAX_ROWS,
        min_rows: int = 0,
        max_rows: int = MAX_ROWS,
        min_columns: int = 0,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL
        ) -> st.SearchStrategy:
    '''
    Return an iterable of TypeBlocks instances, all alligned by row count
    '''
    def constructor(shape: tp.Tuple[int, int]) -> st.SearchStrategy:
        rows, columns = shape
        return st.lists(
                get_type_blocks(
                        min_rows=rows,
                        max_rows=rows,
                        dtype_group=dtype_group
                        ),
                min_size=min_size,
                max_size=max_size
                )

    return get_shape_2d(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns,
            ).flatmap(constructor)

#-------------------------------------------------------------------------------
# index objects

_INDEX_CLS_TO_DEFAULT_DT_GROUP = dict((
            (Index, DTGroup.CORE),
            (IndexGO, DTGroup.CORE),
            (IndexYear, DTGroup.YEAR),
            (IndexYearGO, DTGroup.YEAR),
            (IndexYearMonth, DTGroup.YEAR_MONTH),
            (IndexYearMonthGO, DTGroup.YEAR_MONTH),
            (IndexDate, DTGroup.DATE),
            (IndexDateGO, DTGroup.DATE),
            (IndexHour, DTGroup.HOUR),
            (IndexHourGO, DTGroup.HOUR),
            (IndexMinute, DTGroup.MINUTE),
            (IndexMinuteGO, DTGroup.MINUTE),
            (IndexMillisecond, DTGroup.MILLISECOND),
            (IndexMillisecondGO, DTGroup.MILLISECOND),
            (IndexMicrosecond, DTGroup.MICROSECOND),
            (IndexMicrosecondGO, DTGroup.MICROSECOND),
            (IndexNanosecond, DTGroup.NANOSECOND),
            (IndexNanosecondGO, DTGroup.NANOSECOND),
            ))

def get_index(
        min_size: int = 0,
        max_size: int = MAX_ROWS,
        dtype_group: tp.Optional[DTGroup] = None,
        cls: tp.Type[Index] = Index
        ) -> st.SearchStrategy:
    # NOTE: have observed cases where a non-unqiue index is returned: with float/int 0, or two NaNs. Need to filter
    # using get_labels here forces Index construction from lists, rather than from arrays
    if dtype_group is not None:
        # NOTE: we cannot product if the dtype we get will align with a datetime64 type
        return st.builds(cls, get_array_1d(
                min_size=min_size,
                max_size=max_size,
                unique=True,
                dtype_group=dtype_group
                ))
    return st.builds(cls, get_labels(min_size=min_size, max_size=max_size))

get_index_date: tp.Callable[..., st.SearchStrategy] = partial(get_index,
        cls=IndexDate,
        dtype_group=DTGroup.DATE)
get_index_date.__name__ = 'get_index_date'

get_index_year: tp.Callable[..., st.SearchStrategy] = partial(get_index,
        cls=IndexYear,
        dtype_group=DTGroup.YEAR)
get_index_year.__name__ = 'get_index_year'

get_index_go: tp.Callable[..., st.SearchStrategy] = partial(get_index, cls=IndexGO)
get_index_go.__name__ = 'get_index_go'


def get_index_any(
        min_size: int = 0,
        max_size: int = MAX_ROWS,
        ) -> st.SearchStrategy:

    strategies = []
    for cls, dtype_group in _INDEX_CLS_TO_DEFAULT_DT_GROUP.items():
        st_index = get_index(
                min_size=min_size,
                max_size=max_size,
                dtype_group=dtype_group,
                cls=cls,
                )
        strategies.append(st_index)

    return st.one_of(strategies)


def get_index_hierarchy(
        min_size: int = 1,
        max_size: int = MAX_ROWS,
        min_depth: int = 2,
        max_depth: int = 5,
        dtype_group: tp.Optional[DTGroup] = None,
        cls: tp.Callable[..., IndexHierarchy] = IndexHierarchy.from_labels
        ) -> st.SearchStrategy:

    def constructor(
            labels_spacings: tp.Tuple[tp.Sequence[tp.Sequence[str]], tp.Sequence[tp.Iterable[int]]]
            ) -> st.SearchStrategy:
        # returns an iterable of labels
        labels_proto, spacings = labels_spacings
        depth = len(labels_proto)
        size = len(labels_proto[0])

        # update all labels (except the deepest) by repeating values a number of times, as determined by spacings
        labels: tp.List[tp.Optional[tp.Sequence[str]]] = [None for _ in range(depth)]
        for d in range(depth):
            if d >= depth - 1:
                labels[d] = labels_proto[d]
            else:
                spacing = spacings[d]

                def spans() -> tp.Iterator[tp.Iterator[str]]:
                    idx = 0
                    for count in spacing:
                        if count == 0:
                            continue
                        yield repeat(labels_proto[d][idx], count)
                        idx += count

                labels[d] = list(chain.from_iterable(spans()))

        def label_gen() -> tp.Iterator[tp.List[str]]:
            for i in range(size):
                yield [tp.cast(tp.Sequence[str], labels[d])[i] for d in range(depth)]

        return st.builds(
                cls,
                st.just(label_gen()) # can just handle a generator
                )

    # generate depth-sized lists of candidate leabels and spacings
    def get_labels_spacings(depth_size: tp.Tuple[int, int]) -> st.SearchStrategy:
        depth, size = depth_size

        if dtype_group is not None:
            level = get_array_1d(min_size=size,
                    max_size=size,
                    unique=True,
                    dtype_group=dtype_group)
        else:
            level = get_labels(min_size=size, max_size=size)

        labels = st.lists(level, min_size=depth, max_size=depth)
        # could do depth minus 1
        spacings = st.lists(get_spacing(size), min_size=depth, max_size=depth)
        return st.tuples(labels, spacings).flatmap(constructor)

    # generate depth and size, pass to get get_labels
    return st.tuples(
            st.integers(min_value=min_depth, max_value=max_depth),
            st.integers(min_value=min_size, max_value=max_size)
            ).flatmap(get_labels_spacings)


def get_index_hierarchy_any(
        min_size: int = 1,
        max_size: int = MAX_ROWS,
        min_depth: int = 2,
        max_depth: int = 5,
        ) -> st.SearchStrategy:


    def get_labels_spacings(depth_size: tp.Tuple[int, int]) -> st.SearchStrategy:
        depth, size = depth_size
        args = []
        for _ in range(depth):
            args.append(get_index_any(
                    min_size=size,
                    max_size=size,
                    ))

        return st.one_of(
                st.builds(IndexHierarchy.from_product, *args),
                st.builds(IndexHierarchyGO.from_product, *args),
                )

    return st.tuples(
            st.integers(min_value=min_depth, max_value=max_depth),
            st.integers(min_value=min_size, max_value=max_size)
            ).flatmap(get_labels_spacings)



#-------------------------------------------------------------------------------
# series objects

def get_series(
        min_size: int = 0,
        max_size: int = MAX_ROWS,
        cls: tp.Type[Series] = Series,
        dtype_group: DTGroup = DTGroup.ALL,
        index_cls: tp.Type[Index] = Index,
        index_dtype_group: tp.Optional[DTGroup] = None
        ) -> st.SearchStrategy:

    def constructor(shape: tp.Tuple[int]) -> st.SearchStrategy:
        size = shape[0] # tuple len 1

        if issubclass(index_cls, IndexHierarchy):
            index = get_index_hierarchy(
                    cls=index_cls.from_labels,
                    min_size=size,
                    max_size=size,
                    dtype_group=index_dtype_group,
            )
        else:
            index = get_index(
                    cls=index_cls,
                    min_size=size,
                    max_size=size,
                    dtype_group=index_dtype_group,
            )

        return st.builds(cls,
            get_array_1d(
                    min_size=size,
                    max_size=size,
                    dtype_group=dtype_group
                    ),
            index=index
            )

    return get_shape_1d(min_size=min_size, max_size=max_size).flatmap(constructor)

# label index, values
get_series_date_numeric: tp.Callable[..., st.SearchStrategy] = partial(get_series,
        dtype_group=DTGroup.NUMERIC,
        index_cls=IndexDate,
        index_dtype_group=DTGroup.DATE
        )
get_series_date_numeric.__name__ = 'get_series_date_numeric'

# depth greater than 1 index
get_series_str_dgt1_numeric: tp.Callable[..., st.SearchStrategy] = partial(get_series,
        min_size=1,
        dtype_group=DTGroup.NUMERIC,
        index_cls=IndexHierarchy,
        index_dtype_group=DTGroup.STRING
        )
get_series_str_dgt1_numeric.__name__ = 'get_series_str_dgt1_numeric'

get_series_obj_dgt1_numeric: tp.Callable[..., st.SearchStrategy] = partial(get_series,
        min_size=1,
        dtype_group=DTGroup.NUMERIC,
        index_cls=IndexHierarchy,
        index_dtype_group=DTGroup.OBJECT
        )
get_series_obj_dgt1_numeric.__name__ = 'get_series_obj_dgt1_numeric'


#-------------------------------------------------------------------------------
# frames

def get_frame(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        cls: tp.Type[Frame] = Frame,
        dtype_group: DTGroup = DTGroup.ALL,
        index_cls: tp.Type[Index] = Index,
        index_dtype_group: tp.Optional[DTGroup] = None,
        columns_cls: tp.Type[Index] = Index,
        columns_dtype_group: tp.Optional[DTGroup] = None
        ) -> st.SearchStrategy:

    def constructor(shape: tp.Tuple[int, int]) -> st.SearchStrategy:

        row_count, column_count = shape

        if issubclass(index_cls, IndexHierarchy):
            index = get_index_hierarchy(
                    cls=index_cls.from_labels,
                    min_size=row_count,
                    max_size=row_count,
                    dtype_group=index_dtype_group,
            )
        else:
            index = get_index(
                    cls=index_cls,
                    min_size=row_count,
                    max_size=row_count,
                    dtype_group=index_dtype_group,
            )

        if issubclass(index_cls, IndexHierarchy):
            columns = get_index_hierarchy(
                    cls=columns_cls.from_labels,
                    min_size=column_count,
                    max_size=column_count,
                    dtype_group=columns_dtype_group,
            )
        else:
            columns = get_index(
                    cls=columns_cls,
                    min_size=column_count,
                    max_size=column_count,
                    dtype_group=columns_dtype_group,
            )

        return st.builds(cls,
                get_type_blocks(
                        min_rows=row_count,
                        max_rows=row_count,
                        min_columns=column_count,
                        max_columns=column_count,
                        dtype_group=dtype_group
                        ),
                index=index,
                columns=columns
                )

    return get_shape_2d(min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns
            ).flatmap(constructor)


# label index, columns, values
get_frame_date_str_numeric: tp.Callable[..., st.SearchStrategy] = partial(get_frame,
        dtype_group=DTGroup.NUMERIC,
        index_cls=IndexDate,
        index_dtype_group=DTGroup.DATE,
        columns_cls=Index,
        columns_dtype_group=DTGroup.STRING
        )
get_frame_date_str_numeric.__name__ = 'get_frame_date_str_numeric'

get_frame_go: tp.Callable[..., st.SearchStrategy] = partial(get_frame, cls=FrameGO)
get_frame_go.__name__ = 'get_frame_go'

def get_frame_or_frame_go(
        min_rows: int = 1,
        max_rows: int = MAX_ROWS,
        min_columns: int = 1,
        max_columns: int = MAX_COLUMNS,
        dtype_group: DTGroup = DTGroup.ALL,
        index_cls: tp.Type[Index] = Index,
        index_dtype_group: tp.Optional[DTGroup] = None,
        columns_cls: tp.Type[Index] = Index,
        columns_dtype_group: tp.Optional[DTGroup] = None
        ) -> st.SearchStrategy:
    '''
    Return either a ``Frame`` or a ``FrameGO``,
    '''
    st_frame = get_frame(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns,
            cls=Frame,
            dtype_group=dtype_group,
            index_cls=index_cls,
            index_dtype_group=index_dtype_group,
            columns_cls=columns_cls,
            columns_dtype_group=columns_dtype_group
            )
    st_frame_go = get_frame(
            min_rows=min_rows,
            max_rows=max_rows,
            min_columns=min_columns,
            max_columns=max_columns,
            cls=FrameGO,
            dtype_group=dtype_group,
            index_cls=index_cls,
            index_dtype_group=index_dtype_group,
            columns_cls=columns_cls,
            columns_dtype_group=columns_dtype_group
            )
    return st.one_of((st_frame, st_frame_go))




if __name__ == '__main__':
    import fnmatch
    from argparse import ArgumentParser
    from static_frame.core.display_color import HexColor

    parser = ArgumentParser()
    parser.add_argument('-n', '--name', default=None)
    parser.add_argument('-c', '--count', default=2, type=int)

    options = parser.parse_args()

    local_items = tuple(locals().items())
    for v in (v for k, v in local_items if callable(v) and k.startswith('get')):

        if options.name:
            if not fnmatch.fnmatch(v.__name__, options.name):
                continue

        print(HexColor.format_terminal('grey', '.' * 50))
        print(HexColor.format_terminal('hotpink', str(v.__name__)))

        for x in range(options.count):
            print(HexColor.format_terminal('grey', '.' * 50))
            example = v().example()
            print(repr(example))



import io
import os
import argparse
import typing as tp
import fnmatch
import timeit
from time import sleep
import cProfile
import pstats
import sys
import datetime
import tempfile
from enum import Enum

from pyinstrument import Profiler #type: ignore
from line_profiler import LineProfiler #type: ignore
import gprof2dot #type: ignore

import numpy as np
import pandas as pd
import frame_fixtures as ff
import static_frame as sf

from static_frame.core.display_color import HexColor
from static_frame.core.util import AnyCallable, isin
from static_frame.test.test_case import temp_file


class PerfStatus(Enum):
    EXPLAINED_WIN = (True, True)
    EXPLAINED_LOSS = (True, False)
    UNEXPLAINED_WIN = (False, True)
    UNEXPLAINED_LOSS = (False, False)

    def __str__(self) -> str:
        if self.value[0]:
            v = '✓' # make a check mark
        else:
            v = '?'
        if self.value[1]:
            return HexColor.format_terminal('darkgreen', str(v))
        return HexColor.format_terminal('darkorange', str(v))

class FunctionMetaData(tp.NamedTuple):
    line_target: tp.Optional[AnyCallable] = None
    perf_status: tp.Optional[PerfStatus] = None
    explanation: str = ''

class PerfKey: pass

class Perf(PerfKey):
    NUMBER = 100_000

    def __init__(self) -> None:
        self.meta: tp.Optional[tp.Dict[str, FunctionMetaData]] = None

    def iter_function_names(self, pattern_func: str = '') -> tp.Iterator[str]:
        for name in sorted(dir(self)):
            if name == 'iter_function_names':
                continue
            if pattern_func and not fnmatch.fnmatch(
                    name, pattern_func.lower()):
               continue
            if not name.startswith('_') and callable(getattr(self, name)):
                yield name


class Native(PerfKey): pass
class Reference(PerfKey): pass
class ReferenceMissing(Reference):
    '''For classes that do not / cannot run a reference component.
    '''



#-------------------------------------------------------------------------------

class IndexIterLabelApply(Perf):
    NUMBER = 200

    def __init__(self) -> None:
        super().__init__()


        self.sfi_int = ff.parse('s(100,1)|i(I,int)|c(I,int)').index
        self.pdi_int = self.sfi_int.to_pandas()

class IndexIterLabelApply_N(IndexIterLabelApply, Native):

    def index_int(self) -> None:
        self.sfi_int.iter_label().apply(lambda s: s * 10)

    def index_int_dtype(self) -> None:
        self.sfi_int.iter_label().apply(lambda s: s * 10, dtype=int)

class IndexIterLabelApply_R(IndexIterLabelApply, Reference):

    def index_int(self) -> None:
        # Pandas Index to not have an apply
        pd.Series(self.pdi_int).apply(lambda s: s * 10)

    def index_int_dtype(self) -> None:
        # Pandas Index to not have an apply
        pd.Series(self.pdi_int).apply(lambda s: s * 10)



#-------------------------------------------------------------------------------
class SeriesIsNa(Perf):
    NUMBER = 10_000

    def __init__(self) -> None:
        super().__init__()

        f = ff.parse('s(1000,3)|v(float,object,bool)')
        f = f.assign.loc[(f.index % 12 == 0), 0](np.nan)
        f = f.assign.loc[(f.index % 12 == 0), 1](None)

        self.sfs_float = f.iloc[:, 0]
        self.sfs_object = f.iloc[:, 1]
        self.sfs_bool = f.iloc[:, 2]

        self.pds_float = f.iloc[:, 0].to_pandas()
        self.pds_object = f.iloc[:, 1].to_pandas()
        self.pds_bool = f.iloc[:, 2].to_pandas()

        self.meta = {
            'float_index_auto': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'object_index_auto': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'bool_index_auto': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN, # not copying anything
                ),
            }

class SeriesIsNa_N(SeriesIsNa, Native):

    def float_index_auto(self) -> None:
        self.sfs_float.isna()

    def object_index_auto(self) -> None:
        self.sfs_object.isna()

    def bool_index_auto(self) -> None:
        self.sfs_bool.isna()

class SeriesIsNa_R(SeriesIsNa, Reference):

    def float_index_auto(self) -> None:
        self.pds_float.isna()

    def object_index_auto(self) -> None:
        self.pds_object.isna()

    def bool_index_auto(self) -> None:
        self.pds_bool.isna()

#-------------------------------------------------------------------------------
class SeriesDropNa(Perf):
    NUMBER = 200

    def __init__(self) -> None:
        super().__init__()

        f1 = ff.parse('s(100_000,3)|v(float,object,bool)')
        f1 = f1.assign.loc[(f1.index % 12 == 0), 0](np.nan)
        f1 = f1.assign.loc[(f1.index % 12 == 0), 1](None)

        self.sfs_float_auto = f1.iloc[:, 0]
        self.sfs_object_auto = f1.iloc[:, 1]
        self.sfs_bool_auto = f1.iloc[:, 2]

        self.pds_float_auto = f1.iloc[:, 0].to_pandas()
        self.pds_object_auto = f1.iloc[:, 1].to_pandas()
        self.pds_bool_auto = f1.iloc[:, 2].to_pandas()


        f2 = ff.parse('s(100_000,3)|v(float,object,bool)|i(I,str)|c(I,str)')
        f2 = f2.assign.loc[f2.index.via_str.find('u') >= 0, sf.ILoc[0]](np.nan)
        f2 = f2.assign.loc[f2.index.via_str.find('u') >= 0, sf.ILoc[1]](None)

        self.sfs_float_str = f2.iloc[:, 0]
        self.sfs_object_str = f2.iloc[:, 1]
        self.sfs_bool_str = f2.iloc[:, 2]

        self.pds_float_str = f2.iloc[:, 0].to_pandas()
        self.pds_object_str = f2.iloc[:, 1].to_pandas()
        self.pds_bool_str = f2.iloc[:, 2].to_pandas()

        self.meta = {
            'float_index_auto': FunctionMetaData(
                line_target=sf.Index.__init__,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            'object_index_auto': FunctionMetaData(
                line_target=sf.Series.dropna,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            'bool_index_auto': FunctionMetaData(
                line_target=sf.Series.dropna,
                perf_status=PerfStatus.EXPLAINED_WIN, # not copying anything
                ),

            'float_index_str': FunctionMetaData(
                line_target=sf.Index.__init__,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            'object_index_str': FunctionMetaData(
                line_target=sf.Series.dropna,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            'bool_index_str': FunctionMetaData(
                line_target=sf.Series.dropna,
                perf_status=PerfStatus.EXPLAINED_WIN,
                )
            }

class SeriesDropNa_N(SeriesDropNa, Native):

    def float_index_auto(self) -> None:
        s = self.sfs_float_auto.dropna()
        assert 99999 in s

    def object_index_auto(self) -> None:
        s = self.sfs_object_auto.dropna()
        assert 99999 in s

    def bool_index_auto(self) -> None:
        s = self.sfs_bool_auto.dropna()
        assert 99999 in s


    def float_index_str(self) -> None:
        s = self.sfs_float_str.dropna()
        assert 'zDa2' in s

    def object_index_str(self) -> None:
        s = self.sfs_object_str.dropna()
        assert 'zDa2' in s

    def bool_index_str(self) -> None:
        s = self.sfs_bool_str.dropna()
        assert 'zDa2' in s


class SeriesDropNa_R(SeriesDropNa, Reference):

    def float_index_auto(self) -> None:
        s = self.pds_float_auto.dropna()
        assert 99999 in s

    def object_index_auto(self) -> None:
        s = self.pds_object_auto.dropna()
        assert 99999 in s

    def bool_index_auto(self) -> None:
        s = self.pds_bool_auto.dropna()
        assert 99999 in s


    def float_index_str(self) -> None:
        s = self.pds_float_str.dropna()
        assert 'zDa2' in s

    def object_index_str(self) -> None:
        s = self.pds_object_str.dropna()
        assert 'zDa2' in s

    def bool_index_str(self) -> None:
        s = self.pds_bool_str.dropna()
        assert 'zDa2' in s




#-------------------------------------------------------------------------------
class SeriesFillNa(Perf):
    NUMBER = 100

    def __init__(self) -> None:
        super().__init__()

        f1 = ff.parse('s(100_000,2)|v(float,object)|i(I,str)|c(I,str)')
        f1 = f1.assign.loc[f1.index.via_str.find('u') >= 0, sf.ILoc[0]](np.nan)
        f1 = f1.assign.loc[f1.index.via_str.find('u') >= 0, sf.ILoc[1]](None)

        self.sfs_float_str = f1.iloc[:, 0]
        self.sfs_object_str = f1.iloc[:, 1]

        self.pds_float_str = f1.iloc[:, 0].to_pandas()
        self.pds_object_str = f1.iloc[:, 1].to_pandas()

        from static_frame.core.util import isna_array

        self.meta = {
            'float_index_str': FunctionMetaData(
                line_target=isna_array,
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'object_index_str': FunctionMetaData(
                line_target=isna_array,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                explanation='isna_array does two passes on object arrays',
                ),
            }

class SeriesFillNa_N(SeriesFillNa, Native):

    def float_index_str(self) -> None:
        s = self.sfs_float_str.fillna(0.0)
        assert 'zDa2' in s

    def object_index_str(self) -> None:
        s = self.sfs_object_str.fillna('')
        assert 'zDa2' in s


class SeriesFillNa_R(SeriesFillNa, Reference):

    def float_index_str(self) -> None:
        s = self.pds_float_str.fillna(0.0)
        assert 'zDa2' in s

    def object_index_str(self) -> None:
        s = self.pds_object_str.fillna('')
        assert 'zDa2' in s




#-------------------------------------------------------------------------------
class SeriesDropDuplicated(Perf):
    NUMBER = 500

    def __init__(self) -> None:
        super().__init__()

        f = ff.parse('s(1000,3)|v(float,object,bool)|i(I,str)|c(I,str)')

        self.sfs_float = f.iloc[:, 0]
        self.sfs_float = self.sfs_float.assign.iloc[np.arange(len(self.sfs_float)) % 12 == 0](1.0)
        self.sfs_object = f.iloc[:, 1]
        self.sfs_object = self.sfs_object.assign.iloc[np.arange(len(self.sfs_object)) % 12 == 0](None)
        self.sfs_bool = f.iloc[:, 2]

        self.pds_float = self.sfs_float.to_pandas()
        self.pds_object = self.sfs_object.to_pandas()
        self.pds_bool = self.sfs_bool.to_pandas()


        self.meta = {
            'float_index_str': FunctionMetaData(
                line_target=sf.Index.__init__,
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'object_index_str': FunctionMetaData(
                line_target=sf.Series.drop_duplicated,
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'bool_index_str': FunctionMetaData(
                line_target=sf.Series.drop_duplicated,
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            }

class SeriesDropDuplicated_N(SeriesDropDuplicated, Native):

    def float_index_str(self) -> None:
        s = self.sfs_float.drop_duplicated()
        assert 'zDr0' in s

    def object_index_str(self) -> None:
        s = self.sfs_object.drop_duplicated()
        assert 'zDr0' in s

    def bool_index_str(self) -> None:
        self.sfs_bool.drop_duplicated()

class SeriesDropDuplicated_R(SeriesDropDuplicated, Reference):

    def float_index_str(self) -> None:
        s = self.pds_float.drop_duplicates(keep=False)
        assert 'zDr0' in s

    def object_index_str(self) -> None:
        s = self.pds_object.drop_duplicates(keep=False)
        assert 'zDr0' in s

    def bool_index_str(self) -> None:
        self.pds_bool.drop_duplicates(keep=False)




#-------------------------------------------------------------------------------
class SeriesIterElementApply(Perf):
    NUMBER = 500

    def __init__(self) -> None:
        super().__init__()

        f = ff.parse('s(1000,3)|v(float,object,bool)|i(I,str)|c(I,str)')

        self.sfs_float = f.iloc[:, 0]
        self.sfs_object = f.iloc[:, 1]
        self.sfs_bool = f.iloc[:, 2]

        self.pds_float = f.iloc[:, 0].to_pandas()
        self.pds_object = f.iloc[:, 1].to_pandas()
        self.pds_bool = f.iloc[:, 2].to_pandas()


        from static_frame.core.util import prepare_iter_for_array

        self.meta = {
            'float_index_str': FunctionMetaData(
                line_target=prepare_iter_for_array,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                explanation='prepare_iter_for_array() appears to be the biggest cost'
                ),
            'object_index_str': FunctionMetaData(
                line_target=prepare_iter_for_array,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                explanation='prepare_iter_for_array() appears to be the biggest cost'
                ),
            'bool_index_str': FunctionMetaData(
                line_target=prepare_iter_for_array,
                perf_status=PerfStatus.EXPLAINED_LOSS, # not copying anything
                explanation='prepare_iter_for_array() appears to be the biggest cost'
                ),
            }

class SeriesIterElementApply_N(SeriesIterElementApply, Native):

    def float_index_str(self) -> None:
        self.sfs_float.iter_element().apply(lambda x: str(x))

    def object_index_str(self) -> None:
        self.sfs_object.iter_element().apply(lambda x: str(x))

    def bool_index_str(self) -> None:
        self.sfs_bool.iter_element().apply(lambda x: str(x))

class SeriesIterElementApply_R(SeriesIterElementApply, Reference):

    def float_index_str(self) -> None:
        self.pds_float.apply(lambda x: str(x))

    def object_index_str(self) -> None:
        self.pds_object.apply(lambda x: str(x))

    def bool_index_str(self) -> None:
        self.pds_bool.apply(lambda x: str(x))






#-------------------------------------------------------------------------------
class FrameDropNa(Perf):
    NUMBER = 100

    def __init__(self) -> None:
        super().__init__()

        f1 = ff.parse('s(100,100)|v(float)')
        f1 = f1.assign.loc[(f1.index % 12 == 0),:](np.nan)
        self.sff_float_auto_row = f1
        self.pdf_float_auto_row = f1.to_pandas()

        f2 = ff.parse('s(100,100)|v(float)')
        f2 = f2.assign.loc[:, (f2.columns % 12 == 0)](np.nan)
        self.sff_float_auto_column = f2
        self.pdf_float_auto_column = f2.to_pandas()


        f3 = ff.parse('s(100,100)|v(float)|i(I,str)|c(I,str)')
        f3 = f3.assign.loc[(np.arange(len(f3.index)) % 12 == 0),:](np.nan)
        self.sff_float_str_row = f3
        self.pdf_float_str_row = f3.to_pandas()

        f4 = ff.parse('s(100,100)|v(float)|i(I,str)|c(I,str)')
        f4 = f4.assign.loc[:, (np.arange(len(f4.columns)) % 12 == 0)](np.nan)
        self.sff_float_str_column = f4
        self.pdf_float_str_column = f4.to_pandas()

        self.meta = {
            'float_index_auto_row': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'float_index_auto_column': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'float_index_str_row': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN, # not copying anything
                ),
            'float_index_str_column': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN, # not copying anything
                ),
            }

class FrameDropNa_N(FrameDropNa, Native):

    def float_index_auto_row(self) -> None:
        self.sff_float_auto_row.dropna()

    def float_index_auto_column(self) -> None:
        self.sff_float_auto_column.dropna(axis=1)


    def float_index_str_row(self) -> None:
        self.sff_float_str_row.dropna()

    def float_index_str_column(self) -> None:
        self.sff_float_str_column.dropna(axis=1)


class FrameDropNa_R(FrameDropNa, Reference):

    def float_index_auto_row(self) -> None:
        self.pdf_float_auto_row.dropna()

    def float_index_auto_column(self) -> None:
        self.pdf_float_auto_column.dropna(axis=1)


    def float_index_str_row(self) -> None:
        self.pdf_float_str_row.dropna()

    def float_index_str_column(self) -> None:
        self.pdf_float_str_column.dropna(axis=1)

#-------------------------------------------------------------------------------

class FrameILoc(Perf):

    def __init__(self) -> None:
        super().__init__()

        self.sff1 = ff.parse('s(100,100)')
        self.pdf1 = pd.DataFrame(self.sff1.values)

        self.sff2 = ff.parse('s(100,100)|i(I,str)|c(I,str)')
        self.pdf2 = self.sff2.to_pandas()

        self.meta = {
            'element_index_auto': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'element_index_str': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            }

class FrameILoc_N(FrameILoc, Native):

    def element_index_auto(self) -> None:
        self.sff1.iloc[50, 50]

    def element_index_str(self) -> None:
        self.sff2.iloc[50, 50]

class FrameILoc_R(FrameILoc, Reference):

    def element_index_auto(self) -> None:
        self.pdf1.iloc[50, 50]

    def element_index_str(self) -> None:
        self.pdf2.iloc[50, 50]

#-------------------------------------------------------------------------------

class FrameLoc(Perf):

    def __init__(self) -> None:
        super().__init__()

        self.sff1 = ff.parse('s(100,100)')
        self.pdf1 = pd.DataFrame(self.sff1.values)

        self.sff2 = ff.parse('s(100,100)|i(I,str)|c(I,str)')
        self.pdf2 = self.sff2.to_pandas()

        self.meta = {
            'element_index_auto': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'element_index_str': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            }

class FrameLoc_N(FrameLoc, Native):

    def element_index_auto(self) -> None:
        self.sff1.loc[50, 50]

    def element_index_str(self) -> None:
        self.sff2.loc['zGuv', 'zGuv']

class FrameLoc_R(FrameLoc, Reference):

    def element_index_auto(self) -> None:
        self.pdf1.loc[50, 50]

    def element_index_str(self) -> None:
        self.pdf2.loc['zGuv', 'zGuv']


#-------------------------------------------------------------------------------

class FrameIterSeriesApply(Perf):
    NUMBER = 50

    def __init__(self) -> None:
        super().__init__()


        self.sff_float = ff.parse('s(1000,1000)|i(I,str)|c(I,int)')
        self.pdf_float = self.sff_float.to_pandas()

        self.sff_mixed = ff.parse('s(1000,1000)|v(int,float,bool,str)|i(I,str)|c(I,int)')
        self.pdf_mixed = self.sff_mixed.to_pandas()

        from static_frame.core.type_blocks import TypeBlocks
        from static_frame.core.util import iterable_to_array_1d
        from static_frame.core.util import prepare_iter_for_array

        self.meta = {
            'float_index_str_row': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                line_target=prepare_iter_for_array,
                explanation='appears to all be in apply_iter gen exp'
                ),
            'float_index_str_row_dtype': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            'float_index_str_column': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'float_index_str_column_dtype': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'mixed_index_str_row': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                line_target=TypeBlocks._blocks_to_array,
                explanation='possible improvement with _blocks_to_array in C'
                ),
            'mixed_index_str_row_dtype': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                line_target=iterable_to_array_1d
                ),
            'mixed_index_str_column': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            'mixed_index_str_column_dtype': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                ),
            }

class FrameIterSeriesApply_N(FrameIterSeriesApply, Native):

    def float_index_str_row(self) -> None:
        s = self.sff_float.iter_series(axis=1).apply(lambda s: s.mean())
        assert 'zwVN' in s.index

    def float_index_str_row_dtype(self) -> None:
        s = self.sff_float.iter_series(axis=1).apply(lambda s: s.mean(), dtype=float)
        assert 'zwVN' in s.index


    def float_index_str_column(self) -> None:
        s = self.sff_float.iter_series(axis=0).apply(lambda s: s.mean())
        assert -149082 in s.index

    def float_index_str_column_dtype(self) -> None:
        s = self.sff_float.iter_series(axis=0).apply(lambda s: s.mean(), dtype=float)
        assert -149082 in s.index


    def mixed_index_str_row(self) -> None:
        s = self.sff_mixed.iter_series(axis=1).apply(lambda s: s.iloc[-1])
        assert 'zwVN' in s.index

    def mixed_index_str_row_dtype(self) -> None:
        s = self.sff_mixed.iter_series(axis=1).apply(lambda s: s.iloc[-1], dtype=str)
        assert 'zwVN' in s.index


    def mixed_index_str_column(self) -> None:
        s = self.sff_mixed.iter_series(axis=0).apply(lambda s: s.iloc[-1])
        assert -149082 in s.index

    def mixed_index_str_column_dtype(self) -> None:
        s = self.sff_mixed.iter_series(axis=0).apply(lambda s: s.iloc[-1], dtype=str)
        assert -149082 in s.index



class FrameIterSeriesApply_R(FrameIterSeriesApply, Reference):

    def float_index_str_row(self) -> None:
        s = self.pdf_float.apply(lambda s: s.mean(), axis=1)
        assert 'zwVN' in s.index

    def float_index_str_row_dtype(self) -> None:
        s = self.pdf_float.apply(lambda s: s.mean(), axis=1)
        assert 'zwVN' in s.index


    def float_index_str_column(self) -> None:
        s = self.pdf_float.apply(lambda s: s.mean(), axis=0)
        assert -149082 in s.index

    def float_index_str_column_dtype(self) -> None:
        s = self.pdf_float.apply(lambda s: s.mean(), axis=0)
        assert -149082 in s.index


    def mixed_index_str_row(self) -> None:
        s = self.pdf_mixed.apply(lambda s: s.iloc[-1], axis=1)
        assert 'zwVN' in s.index

    def mixed_index_str_row_dtype(self) -> None:
        s = self.pdf_mixed.apply(lambda s: s.iloc[-1], axis=1)
        assert 'zwVN' in s.index


    def mixed_index_str_column(self) -> None:
        s = self.pdf_mixed.apply(lambda s: s.iloc[-1], axis=0)
        assert -149082 in s.index

    def mixed_index_str_column_dtype(self) -> None:
        s = self.pdf_mixed.apply(lambda s: s.iloc[-1], axis=0)
        assert -149082 in s.index

#-------------------------------------------------------------------------------
class FrameIterGroupApply(Perf):
    NUMBER = 1000

    def __init__(self) -> None:
        super().__init__()

        self.sff_int_index_str = ff.parse('s(1000,10)|v(int)|i(I,str)|c(I,str)').assign[sf.ILoc[0]].apply(lambda s: s % 10).assign[sf.ILoc[1]].apply(lambda s: s % 2)
        self.pdf_int_index_str = self.sff_int_index_str.to_pandas()


        self.sff_str_index_str = ff.parse('s(1000,10)|v(str)|i(I,str)|c(I,str)').assign[
                sf.ILoc[0]].apply(lambda s: s.iter_element().apply(
                        lambda e: chr(ord(e[3]) % 10 + 97))).assign[
                sf.ILoc[1]].apply(lambda s: s.iter_element().apply(
                        lambda e: chr(ord(e[3]) % 2 + 97)))

        self.pdf_str_index_str = self.sff_str_index_str.to_pandas()


        from static_frame.core.type_blocks import TypeBlocks
        # from static_frame.core.util import iterable_to_array_1d
        # from static_frame.core.util import prepare_iter_for_array

        self.meta = {
            'int_index_str_double': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            }

class FrameIterGroupApply_N(FrameIterGroupApply, Native):

    def int_index_str_single(self) -> None:
        self.sff_int_index_str.iter_group('zZbu').apply(lambda f: len(f))

    def int_index_str_double(self) -> None:
        self.sff_int_index_str.iter_group(['zZbu', 'ztsv']).apply(lambda f: len(f))


    def str_index_str_single(self) -> None:
        self.sff_str_index_str.iter_group('zZbu').apply(lambda f: len(f))

    def str_index_str_double(self) -> None:
        self.sff_str_index_str.iter_group(['zZbu', 'ztsv']).apply(lambda f: len(f))


class FrameIterGroupApply_R(FrameIterGroupApply, Reference):

    def int_index_str_single(self) -> None:
        self.pdf_int_index_str.groupby('zZbu').apply(lambda f: len(f))

    def int_index_str_double(self) -> None:
        # NOTE: this produces a hierarchical index
        self.pdf_int_index_str.groupby(['zZbu', 'ztsv']).apply(lambda f: len(f))


    def str_index_str_single(self) -> None:
        self.pdf_str_index_str.groupby('zZbu').apply(lambda f: len(f))

    def str_index_str_double(self) -> None:
        # NOTE: this produces a hierarchical index
        self.pdf_str_index_str.groupby(['zZbu', 'ztsv']).apply(lambda f: len(f))


#-------------------------------------------------------------------------------
class Pivot(Perf):
    NUMBER = 150

    def __init__(self) -> None:
        super().__init__()

        self.sff1 = ff.parse('s(100_000,10)|v(int,str,bool)|c(I,str)|i(I,int)')
        self.pdf1 = self.sff1.to_pandas()

        # narrow eav table
        self.sff2 = ff.parse('s(100_000,3)|v(int,int,int)').assign[0].apply(
                lambda s: s % 6).assign[1].apply(
                lambda s: s % 12
                )
        self.pdf2 = self.sff2.to_pandas()

        # more than one data fields
        self.sff3 = ff.parse('s(100_000,4)|v(int,int,int,int)').assign[0].apply(
                lambda s: s % 6).assign[1].apply(
                lambda s: s % 12
                )

        from static_frame.core.pivot import pivot_outer_index
        from static_frame import TypeBlocks
        from static_frame.core.util import array_to_groups_and_locations
        self.meta = {
            'index1_columns0_data2': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_WIN,
                line_target=array_to_groups_and_locations,
                ),
            'index1_columns1_data1': FunctionMetaData(
                line_target=pivot_outer_index,
                perf_status=PerfStatus.EXPLAINED_LOSS,
                ),
            }

class Pivot_N(Pivot, Native):

    def index1_columns0_data2(self) -> None:
        post = self.sff1.pivot(index_fields='zUvW', data_fields=('zZbu', 'zkuW'))
        assert post.shape == (2, 2)

    def index1_columns1_data1(self) -> None:
        post = self.sff2.pivot(index_fields=0, columns_fields=1)
        assert post.shape == (6, 12)


class Pivot_R(Pivot, Reference):

    def index1_columns0_data2(self) -> None:
        post = self.pdf1.pivot_table(index='zUvW', values=('zZbu', 'zkuW'), aggfunc=np.nansum)
        assert post.shape == (2, 2)

    def index1_columns1_data1(self) -> None:
        post = self.pdf2.pivot_table(index=0, columns=1, aggfunc=np.nansum)
        assert post.shape == (6, 12)

#-------------------------------------------------------------------------------
class BusItemsZipPickle(Perf):
    NUMBER = 1

    def __init__(self) -> None:
        super().__init__()

        def items() -> tp.Iterator[tp.Tuple[str, sf.Frame]]:
            f = ff.parse(f's(2,2)|v(int)|i(I,str)|c(I,str)')
            for i in range(10_000):
                yield str(i), f

        frames = sf.Series.from_items(items(), dtype=object)
        _, self.fp = tempfile.mkstemp(suffix='.zip')
        b1 = sf.Bus(frames)
        b1.to_zip_pickle(self.fp)

        # self.meta = {
        #     'int_index_str_double': FunctionMetaData(
        #         perf_status=PerfStatus.EXPLAINED_LOSS,
        #         None
        #         ),
        #     }

    def __del__(self) -> None:
        os.unlink(self.fp)

class BusItemsZipPickle_N(BusItemsZipPickle, Native):

    def int_index_str(self) -> None:
        bus = sf.Bus.from_zip_pickle(self.fp, max_persist=100)
        for label, frame in bus.items():
           assert frame.shape[0] == 2

class BusItemsZipPickle_R(BusItemsZipPickle, ReferenceMissing):

    def int_index_str(self) -> None:
        pass

#-------------------------------------------------------------------------------
class FrameToParquet(Perf):
    NUMBER = 4

    def __init__(self) -> None:
        super().__init__()
        _, self.fp = tempfile.mkstemp(suffix='.zip')

        self.sff1 = ff.parse('s(10,10_000)|v(int,int,bool,float,float)|i(I,str)|c(I,str)')
        self.pdf1 = self.sff1.to_pandas()

        self.sff2 = ff.parse('s(10_000,10)|v(int,int,bool,float,float)|i(I,str)|c(I,str)')
        self.pdf2 = self.sff2.to_pandas()


        # self.meta = {
        #     'int_index_str_double': FunctionMetaData(
        #         perf_status=PerfStatus.EXPLAINED_LOSS,
        #         None
        #         ),
        #     }

    def __del__(self) -> None:
        os.unlink(self.fp)

class FrameToParquet_N(FrameToParquet, Native):

    def write_wide_mixed_index_str(self) -> None:
        self.sff1.to_parquet(self.fp)

    def write_tall_mixed_index_str(self) -> None:
        self.sff2.to_parquet(self.fp)


class FrameToParquet_R(FrameToParquet, Reference):

    def write_wide_mixed_index_str(self) -> None:
        self.pdf1.to_parquet(self.fp)

    def write_tall_mixed_index_str(self) -> None:
        self.pdf2.to_parquet(self.fp)



#-------------------------------------------------------------------------------
class FrameToNPZ(Perf):
    NUMBER = 1

    def __init__(self) -> None:
        super().__init__()
        _, self.fp = tempfile.mkstemp(suffix='.zip')

        self.sff1 = ff.parse('s(10,10_000)|v(int,int,bool,float,float)|i(I,str)|c(I,str)')
        self.pdf1 = self.sff1.to_pandas()


        # self.meta = {
        #     'int_index_str_double': FunctionMetaData(
        #         perf_status=PerfStatus.EXPLAINED_LOSS,
        #         None
        #         ),
        #     }

    def __del__(self) -> None:
        os.unlink(self.fp)

class FrameToNPZ_N(FrameToNPZ, Native):

    def wide_mixed_index_str(self) -> None:
        self.sff1.to_npz(self.fp)

class FrameToNPZ_R(FrameToNPZ, Reference):

    # NOTE: benchmark is SF to_parquet
    def wide_mixed_index_str(self) -> None:
        self.sff1.to_parquet(self.fp)

#-------------------------------------------------------------------------------
class Group(Perf):
    NUMBER = 150

    def __init__(self) -> None:
        super().__init__()

        self.sff1 = ff.parse('s(100_000,10)|v(int,str,bool)|c(I,str)|i(I,int)')
        self.pdf1 = self.sff1.to_pandas()

        # narrow eav table
        self.sff2 = ff.parse('s(100_000,3)|v(int,int,int)').assign[0].apply(
                lambda s: s % 6).assign[1].apply(
                lambda s: s % 100
                )
        self.pdf2 = self.sff2.to_pandas()

        from static_frame import Frame
        from static_frame import TypeBlocks
        from static_frame.core.util import array_to_groups_and_locations
        self.meta = {
            'wide_group_2': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                line_target=Frame._axis_group_iloc_items,
                explanation='nearly identical, favoring slower'
                ),
            'tall_group_100': FunctionMetaData(
                perf_status=PerfStatus.EXPLAINED_LOSS,
                line_target=Frame._axis_group_iloc_items,
                ),
            }

class Group_N(Group, Native):

    def wide_group_2(self) -> None:
        post = tuple(self.sff1.iter_group_items('zUvW'))
        assert len(post) == 2

    def tall_group_100(self) -> None:
        post = tuple(self.sff2.iter_group_items(1))
        assert len(post) == 100


class Group_R(Group, Reference):

    def wide_group_2(self) -> None:
        post = tuple(self.pdf1.groupby('zUvW'))
        assert len(post) == 2

    def tall_group_100(self) -> None:
        post = tuple(self.pdf2.groupby(1))
        assert len(post) == 100




#-------------------------------------------------------------------------------
class FrameFromConcat(Perf):
    NUMBER = 50

    def __init__(self) -> None:
        super().__init__()
        self.tall_mixed_sff1 = [
            ff.parse('s(10_000,10)|v(int,str,int,bool)')
            for _ in range(20)
            ]
        self.tall_mixed_pdf1 = [f.to_pandas() for f in self.tall_mixed_sff1]


        self.tall_uniform_sff1 = [
            ff.parse('s(10_000,10)|v(float)')
            for _ in range(20)
            ]
        self.tall_uniform_pdf1 = [f.to_pandas() for f in self.tall_uniform_sff1]


        from static_frame import Frame
        # from static_frame import TypeBlocks
        # from static_frame.core.util import array_to_groups_and_locations
        # self.meta = {
        #     'tall_uniform_20': FunctionMetaData(
        #         # perf_status=PerfStatus.EXPLAINED_LOSS,
        #         line_target=Frame.from_concat,
        #         # explanation='nearly identical, favoring slower'
        #         ),
        #     }

class FrameFromConcat_N(FrameFromConcat, Native):

    def tall_mixed_20(self) -> None:
        f = sf.Frame.from_concat(self.tall_mixed_sff1, index=sf.IndexAutoFactory)
        assert f.shape == (200_000, 10)

    def tall_uniform_20(self) -> None:
        f = sf.Frame.from_concat(self.tall_uniform_sff1, index=sf.IndexAutoFactory)
        assert f.shape == (200_000, 10)

class FrameFromConcat_R(FrameFromConcat, Reference):

    def tall_mixed_20(self) -> None:
        df = pd.concat(self.tall_mixed_pdf1)
        assert df.shape == (200_000, 10)

    def tall_uniform_20(self) -> None:
        df = pd.concat(self.tall_uniform_pdf1)
        assert df.shape == (200_000, 10)


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------

def get_arg_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(
            description='Performance testing and profiling',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog='''Example:

Performance comparison of all dropna tests:

python3 test_performance.py '*dropna' --performance

Profiling outpout for static-frame dropna:

python3 test_performance.py SeriesIntFloat_dropna --profile
            '''
            )
    p.add_argument('patterns',
            help='Names of classes to match using fn_match syntax',
            nargs='+',
            )
    # p.add_argument('--modules',
    #         help='Names of modules to find tests',
    #         nargs='+',
    #         default=('core',),
    #         )
    p.add_argument('--profile',
            help='Turn on profiling with cProfile',
            action='store_true',
            default=False,
            )
    p.add_argument('--graph',
            help='Produce a call graph of cProfile output',
            action='store_true',
            default=False,
            )
    p.add_argument('--instrument',
            help='Turn on instrumenting with pyinstrument',
            action='store_true',
            default=False,
            )
    p.add_argument('--performance',
            help='Turn on performance measurements',
            action='store_true',
            default=False,
            )
    p.add_argument('--line',
            help='Turn on line profiler',
            action='store_true',
            default=False,
            )
    return p


def yield_classes(
        pattern: str
        ) -> tp.Iterator[
                tp.Tuple[
                    tp.Dict[tp.Type[PerfKey], tp.Type[PerfKey]],
                    str]]:

    if '.' in pattern:
        pattern_cls, pattern_func = pattern.split('.')
    else:
        pattern_cls, pattern_func = pattern, '*'

    for cls_perf in Perf.__subclasses__(): # only get one level
        if pattern_cls and not fnmatch.fnmatch(
                cls_perf.__name__.lower(), pattern_cls.lower()):
            continue

        runners: tp.Dict[tp.Type[PerfKey], tp.Type[PerfKey]] = {Perf: cls_perf}

        for cls_runner in cls_perf.__subclasses__():
            for cls in (Native, Reference):
                if issubclass(cls_runner, cls):
                    runners[cls] = cls_runner
                    break
        yield runners, pattern_func


def profile(
        cls_runner: tp.Type[Perf],
        pattern_func: str,
        ) -> None:
    '''
    Profile the `sf` function from the supplied class.
    '''
    runner = cls_runner()
    for name in runner.iter_function_names(pattern_func):
        f = getattr(runner, name)
        pr = cProfile.Profile()

        pr.enable()
        for _ in range(runner.NUMBER):
            f()
        pr.disable()

        s = io.StringIO()
        ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
        ps.print_stats()
        print(s.getvalue())

def graph(
        cls_runner: tp.Type[Perf],
        pattern_func: str,
        threshold_edge: float = 0.1,
        threshold_node: float = 0.5,
        ) -> None:
    '''
    Profile the `sf` function from the supplied class.
    '''
    runner = cls_runner()
    for name in runner.iter_function_names(pattern_func):
        _, fp = tempfile.mkstemp(suffix='', text=True)
        fp_pstat = fp + '.pstat'
        fp_dot = fp + '.dot'
        fp_png = fp + '.png'

        f = getattr(runner, name)
        pr = cProfile.Profile()

        pr.enable()
        for _ in range(runner.NUMBER):
            f()
        pr.disable()

        ps = pstats.Stats(pr)
        ps.dump_stats(fp_pstat)

        gprof2dot.main([
            '--format', 'pstats',
            '--output', fp_dot,
            '--edge-thres', threshold_edge, # 0.1 default
            '--node-thres', threshold_node, # 0.5 default
            fp_pstat
        ])
        os.system(f'dot {fp_dot} -Tpng -Gdpi=300 -o {fp_png}; eog {fp_png} &')

def instrument(
        cls_runner: tp.Type[Perf],
        pattern_func: str,
        timeline: bool = False,
        ) -> None:
    '''
    Profile the `sf` function from the supplied class.
    '''
    runner = cls_runner()
    for name in runner.iter_function_names(pattern_func):
        f = getattr(runner, name)
        profiler = Profiler(interval=0.0001) # default is 0.001, 1 ms

        if timeline:
            profiler.start()
            f()
            profiler.stop()
        else:
            profiler.start()
            for _ in range(runner.NUMBER):
                f()
            profiler.stop()

        print(profiler.output_text(unicode=True, color=True, timeline=timeline, show_all=True))

def line(
        cls_runner: tp.Type[Perf],
        pattern_func: str,
        ) -> None:
    runner = cls_runner()
    for name in runner.iter_function_names(pattern_func):
        f = getattr(runner, name)
        profiler = LineProfiler()
        if not runner.meta:
            raise NotImplementedError('must define runner.meta')
        profiler.add_function(runner.meta[name].line_target)
        profiler.enable()
        f()
        profiler.disable()
        profiler.print_stats()

#-------------------------------------------------------------------------------

PerformanceRecord = tp.MutableMapping[str,
        tp.Union[str, float, bool, tp.Optional[PerfStatus]]]

def performance(
        bundle: tp.Dict[tp.Type[PerfKey], tp.Type[PerfKey]],
        pattern_func: str,
        ) -> tp.Iterator[PerformanceRecord]:

    cls_perf = bundle[Perf]
    assert issubclass(cls_perf, Perf)

    cls_native = bundle[Native]
    cls_reference = bundle[Reference]

    # TODO: check native/ref have the same  iterations
    runner_n = cls_native()
    runner_r = cls_reference()
    assert isinstance(runner_n, Perf)

    for func_name in runner_n.iter_function_names(pattern_func):
        row: PerformanceRecord = {}
        row['name'] = f'{cls_perf.__name__}.{func_name}'
        row['iterations'] = cls_perf.NUMBER

        for label, runner in ((Native, runner_n), (Reference, runner_r)):
            if isinstance(runner, ReferenceMissing):
                row[label.__name__] = np.nan
            else:
                row[label.__name__] = timeit.timeit(
                        f'runner.{func_name}()',
                        globals=locals(),
                        number=cls_perf.NUMBER)

        row['n/r'] = row[Native.__name__] / row[Reference.__name__] #type: ignore
        row['r/n'] = row[Reference.__name__] / row[Native.__name__] #type: ignore
        row['win'] = row['r/n'] > .99 if not np.isnan(row['r/n']) else True #type: ignore

        if runner_n.meta is not None and func_name in runner_n.meta:
            row['status'] = runner_n.meta[func_name].perf_status
            row['explanation'] = runner_n.meta[func_name].explanation
        else:
            row['status'] = (PerfStatus.UNEXPLAINED_WIN if row['win']
                    else PerfStatus.UNEXPLAINED_LOSS)
            row['explanation'] = ''

        yield row


def performance_tables_from_records(
        records: tp.Iterable[PerformanceRecord],
        ) -> tp.Tuple[sf.Frame, sf.Frame]:

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))
        if isinstance(v, str):
            if key[1] == 'explanation':
                return HexColor.format_terminal('gray', v)
            if key[1] == 'name':
                name_root = v.split('.')[0]
                if name_root != name_root_last:
                    name_root_last = name_root
                    name_root_count += 1
                if name_root_count % 2:
                    return HexColor.format_terminal('lavender', v)
                return HexColor.format_terminal('lightslategrey', v)
        return str(v)

    frame = sf.FrameGO.from_dict_records(records)

    fields = ['Native', 'Reference', 'n/r', 'r/n']
    stats = sf.Frame.from_concat((
            frame[fields].min().rename('min'),
            frame[fields].max().rename('max'),
            frame[fields].mean().rename('mean'),
            frame[fields].median().rename('median'),
            frame[fields].std(ddof=1).rename('std')
            )).rename(index='name').unset_index()
    # import ipdb; ipdb.set_trace()
    composit = sf.Frame.from_concat((frame, stats), columns=frame.columns, index=sf.IndexAutoFactory)
    display = composit.iter_element_items().apply(format)
    # display = display[display.columns.drop.loc['status'].values.tolist() + ['status']]
    # display = display[[c for c in display.columns if '/' not in c]]
    return frame, display

def main() -> None:

    options = get_arg_parser().parse_args()
    records: tp.List[PerformanceRecord] = []

    for pattern in options.patterns:
        for bundle, pattern_func in yield_classes(pattern):
            if options.performance:
                records.extend(performance(bundle, pattern_func))
            if options.profile:
                profile(bundle[Native], pattern_func) #type: ignore
            if options.graph:
                graph(bundle[Native], pattern_func) #type: ignore
            if options.instrument:
                instrument(bundle[Native], pattern_func) #type: ignore
            if options.line:
                line(bundle[Native], pattern_func) #type: ignore

    itemize = False # make CLI option maybe

    if records:

        from static_frame import DisplayConfig

        print(str(datetime.datetime.now()))

        pairs = []
        pairs.append(('python', sys.version.split(' ')[0]))
        for package in (np, pd, sf):
            pairs.append((package.__name__, package.__version__))
        print('|'.join(':'.join(pair) for pair in pairs))

        frame, display = performance_tables_from_records(records)

        config = DisplayConfig(
                cell_max_width_leftmost=np.inf,
                cell_max_width=np.inf,
                type_show=False,
                display_rows=200,
                include_index=False,
                )
        print(display.display(config))

        if itemize:
            alt = display.T
            for c in alt.columns:
                print(c)
                print(alt[c].sort_values().display(config))

        # import ipdb; ipdb.set_trace()
        # if 'sf/pd' in frame.columns:
        #     print('mean: {}'.format(round(frame['sf/pd'].mean(), 6)))
        #     print('wins: {}/{}'.format((frame['sf/pd'] < 1.05).sum(), len(frame)))



if __name__ == '__main__':
    main()




import typing as tp
import unittest

import numpy as np

from hypothesis import strategies as st
from hypothesis import given

from static_frame.test.property import strategies as sfst

from static_frame.test.test_case import TestCase

from static_frame import TypeBlocks
# from static_frame import Index
# from static_frame import IndexHierarchy
# from static_frame import Series
# from static_frame import Frame


class TestUnit(TestCase):


    @given(st.lists(sfst.get_shape_2d(), min_size=1), sfst.get_labels(min_size=1))
    def test_from_element_items(self,
            shapes: tp.List[tp.Tuple[int, int]],
            labels: tp.Sequence[tp.Hashable]
            ) -> None:

        # use shapes to get coordinates, where the max shape + 1 is the final shape
        shape = tuple(np.array(shapes).max(axis=0) + 1)

        def values() -> tp.Iterator[tp.Tuple[tp.Tuple[int, int], tp.Hashable]]:
            for idx, coord in enumerate(shapes):
                yield coord, labels[idx % len(labels)]

        post = TypeBlocks.from_element_items(values(), shape=shape, dtype=object)
        self.assertEqual(post.shape, shape)


    @given(st.integers(max_value=sfst.MAX_COLUMNS))
    def test_from_zero_size_shape(self, value: int) -> None:

        for shape in ((0, value), (value, 0)):
            post = TypeBlocks.from_zero_size_shape(shape=shape)
            self.assertEqual(post.shape, shape)


    @given(sfst.get_type_blocks())
    def test_basic_attributes(self, tb: TypeBlocks) -> None:
        self.assertEqual(len(tb.dtypes), tb.shape[1])
        self.assertEqual(len(tb.shapes), len(tb.mloc))
        self.assertEqual(tb.copy().shape, tb.shape)
        self.assertEqual(tb.ndim, 2)
        self.assertEqual(tb.unified, len(tb.mloc) <= 1)

        if tb.shape[0] > 0 and tb.shape[1] > 0:
            self.assertTrue(tb.size > 0)
            self.assertTrue(tb.nbytes > 0)
        else:
            self.assertTrue(tb.size == 0)
            self.assertTrue(tb.nbytes == 0)



    @given(sfst.get_type_blocks())
    def test_values(self, tb: TypeBlocks) -> None:
        values = tb.values
        self.assertEqual(values.shape, tb.shape)
        self.assertEqual(values.dtype, tb._row_dtype)


    @given(sfst.get_type_blocks())
    def test_axis_values(self, tb: TypeBlocks) -> None:
        # this test found a flaw in axis_values when dealing with axis 1 and unified,  1D type blocks
        for axis in (0, 1):
            for reverse in (True, False):
                post = tuple(tb.axis_values(axis=axis, reverse=reverse))
                for idx, array in enumerate(post):
                    self.assertTrue(len(array) == tb.shape[axis])
                    if axis == 0 and not reverse: # colums
                        self.assertTrue(array.dtype == tb.dtypes[idx])
                    elif axis == 0 and reverse: # colums
                        self.assertTrue(array.dtype == tb.dtypes[tb.shape[1] - 1 - idx])
                    else:
                        # NOTE: only checking kinde because found cases where byte-order deviates
                        self.assertTrue(array.dtype.kind == tb._row_dtype.kind)


    @given(sfst.get_type_blocks())
    def test_element_items(self, tb: TypeBlocks) -> None:
        # NOTE: this found a flaw in _extract_iloc where we tried to optimize selection with a unified array
        count = 0
        for k, v in tb.element_items():
            count += 1
            v_extract = tb.iloc[k]
            self.assertEqualWithNaN(v, v_extract)
        self.assertEqual(count, tb.size)

    @given(sfst.get_type_blocks())
    def test_reblock_signature(self, tb: TypeBlocks) -> None:
        post = tuple(tb._reblock_signature())
        unique_dtypes = np.unique(tb.dtypes)
        # the reblock signature must be have at least as many entries as types
        self.assertTrue(len(post) >= len(unique_dtypes))
        # sum of column widths is qual to columns in shape
        self.assertTrue(sum(p[1] for p in post), tb.shape[1])


    @given(sfst.get_type_blocks(), sfst.get_type_blocks())
    def test_block_compatible(self, tb1: TypeBlocks, tb2: TypeBlocks) -> None:

        for axis in (None, 0, 1):
            post1 = tb1.block_compatible(tb2, axis)
            post2 = tb2.block_compatible(tb1, axis)
            # either direction gets the same result
            self.assertTrue(post1 == post2)
            # if the shapes are different, they cannot be block compatible
            if axis is None and tb1.shape != tb2.shape:
                self.assertFalse(post1)


    @given(sfst.get_type_blocks(), sfst.get_type_blocks())
    def test_reblock_compatible(self, tb1: TypeBlocks, tb2: TypeBlocks) -> None:

        post1 = tb1.reblock_compatible(tb2)
        post2 = tb2.reblock_compatible(tb1)
        # either direction gets the same result
        self.assertTrue(post1 == post2)
        # if the shapes are different, they cannot be block compatible
        if tb1.shape[1] != tb2.shape[1]:
            self.assertFalse(post1)

    @unittest.skip('pending')
    def test_concatenate_blocks(self) -> None:
        pass

    @given(sfst.get_type_blocks())
    def test_consolidate_blocks(self, tb: TypeBlocks) -> None:

        tb_post = TypeBlocks.from_blocks(tb.consolidate_blocks(tb._blocks))
        self.assertEqual(tb_post.shape, tb.shape)
        self.assertTrue((tb_post.dtypes == tb.dtypes).all())

    @given(sfst.get_type_blocks())
    def test_reblock(self, tb: TypeBlocks) -> None:
        tb_post = TypeBlocks.from_blocks(tb._reblock())
        self.assertEqual(tb_post.shape, tb.shape)
        self.assertTrue((tb_post.dtypes == tb.dtypes).all())

    @given(sfst.get_type_blocks())
    def test_consolidate(self, tb: TypeBlocks) -> None:
        tb_post = tb.consolidate()
        self.assertEqual(tb_post.shape, tb.shape)
        self.assertTrue((tb_post.dtypes == tb.dtypes).all())

    @unittest.skip('pending')
    def test_resize_blocks(self) -> None:
        pass

    @unittest.skip('pending')
    def test_group(self) -> None:
        pass

    @unittest.skip('pending')
    def test_ufunc_axis_skipna(self) -> None:
        pass

    @given(sfst.get_type_blocks())
    def test_display(self, tb: TypeBlocks) -> None:
        post = tb.display()
        self.assertTrue(len(post) > 0)

    @unittest.skip('pending')
    def test_cols_to_slice(self) -> None:
        pass

    @unittest.skip('pending')
    def test_indices_to_contiguous_pairs(self) -> None:
        pass

    @unittest.skip('pending')
    def test_key_to_block_slices(self) -> None:
        pass

    @unittest.skip('pending')
    def test_mask_blocks(self) -> None:
        pass

    @unittest.skip('pending')
    def test_astype_blocks(self) -> None:
        pass

    @unittest.skip('pending')
    def test_shift_blocks(self) -> None:
        pass

    @given(sfst.get_type_blocks())
    def test_assign_blocks_from_keys(self, tb1: TypeBlocks) -> None:

        # assigning a single value from a list of column keys
        for i in range(tb1.shape[1]):
            tb2 = TypeBlocks.from_blocks(tb1._assign_from_iloc_by_unit(
                    column_key=[i], value=300))
            self.assertTrue(tb1.shape == tb2.shape)
            # no more than one type should be changed
            self.assertTrue((tb1.dtypes != tb2.dtypes).sum() <= 1)

        # assigning a single value from a list of row keys
        for i in range(tb1.shape[0]):
            tb3 = TypeBlocks.from_blocks(tb1._assign_from_iloc_by_unit(
                    row_key=[i], value=300))
            self.assertTrue(tb1.shape == tb3.shape)
            self.assertTrue(tb3.iloc[i, 0] == 300)

        # column slices to the end
        for i in range(tb1.shape[1]):
            tb4 = TypeBlocks.from_blocks(tb1._assign_from_iloc_by_unit(
                    column_key=slice(i, None), value=300))
            self.assertTrue(tb1.shape == tb4.shape)
            # we have as many or more blocks
            self.assertTrue(len(tb4.shapes) >= len(tb1.shapes))


    @unittest.skip('pending')
    def test_assign_blocks_from_boolean_blocks(self) -> None:
        pass

    @unittest.skip('pending')
    def test_slice_blocks(self) -> None:
        pass

    @unittest.skip('pending')
    def test_extract_array(self) -> None:
        pass

    @unittest.skip('pending')
    def test_extract(self) -> None:
        pass

    @unittest.skip('pending')
    def test_extract_iloc(self) -> None:
        pass

    @unittest.skip('pending')
    def test_extract_iloc_mask(self) -> None:
        pass

    @unittest.skip('pending')
    def test_extract_iloc_assign(self) -> None:
        pass

    @given(sfst.get_type_blocks(min_rows=1, min_columns=1))
    def test_drop(self, tb: TypeBlocks) -> None:

        for row in range(tb.shape[0]):
            tb_post1 = tb.drop(row)
            self.assertTrue(tb_post1.shape[0] == tb.shape[0] - 1)

        if tb.shape[0] > 2:
            for start in range(1, tb.shape[0]):
                tb_post2 = tb.drop(slice(start, None))
                self.assertTrue(tb_post2.shape[0] == start)

        for col in range(tb.shape[1]):
            tb_post3 = tb.drop((None, col))
            self.assertTrue(tb_post3.shape[1] == tb.shape[1] - 1)

        if tb.shape[1] > 2:
            for start in range(1, tb.shape[1]):
                tb_post4 = tb.drop((None, slice(start, None)))
                self.assertTrue(tb_post4.shape[1] == start)



    @unittest.skip('pending')
    def test_getitem(self) -> None:
        pass

    @unittest.skip('pending')
    def test_ufunc_unary_operator(self) -> None:
        pass

    @unittest.skip('pending')
    def test_block_shape_slices(self) -> None:
        pass

    @unittest.skip('pending')
    def test_ufunc_binary_operator(self) -> None:
        pass

    @unittest.skip('pending')
    def test_transpose(self) -> None:
        pass

    @unittest.skip('pending')
    def test_isna(self) -> None:
        pass

    @unittest.skip('pending')
    def test_notna(self) -> None:
        pass

    @unittest.skip('pending')
    def test_fillna_leading(self) -> None:
        pass

    @unittest.skip('pending')
    def test_fillna_trailing(self) -> None:
        pass

    @unittest.skip('pending')
    def test_fillna_forward(self) -> None:
        pass

    @unittest.skip('pending')
    def test_fillna_backward(self) -> None:
        pass

    @unittest.skip('pending')
    def test_dropna_to_keep_locations(self) -> None:
        pass

    @unittest.skip('pending')
    def test_fillna(self) -> None:
        pass


    @given(sfst.get_type_blocks_aligned_array())
    def test_append(self, tb_aligned_array: tp.Tuple[TypeBlocks, np.ndarray]) -> None:
        tb, aa = tb_aligned_array
        shape_original = tb.shape
        tb.append(aa)
        if aa.ndim == 1:
            self.assertEqual(tb.shape[1], shape_original[1] + 1)
        else:
            self.assertEqual(tb.shape[1], shape_original[1] + aa.shape[1])

    @given(sfst.get_type_blocks_aligned_type_blocks(min_size=2, max_size=2))
    def test_extend(self, tbs: tp.Sequence[TypeBlocks]) -> None:
        front = tbs[0]
        back = tbs[1]
        shape_original = front.shape
        # extend with type blocks
        front.extend(back)
        self.assertEqual(front.shape,
                (shape_original[0], shape_original[1] + back.shape[1]))

        # extend with iterable of arrays
        front.extend(back._blocks)
        self.assertEqual(front.shape,
                (shape_original[0], shape_original[1] + back.shape[1] * 2))


if __name__ == '__main__':
    unittest.main()



import os
from sphinx.cmd.build import main

if __name__ == '__main__':
    doc_dir = os.path.abspath(os.path.dirname(__file__))
    doctrees_dir = os.path.join(doc_dir, 'build', 'doctrees')
    source_dir = os.path.join(doc_dir, 'source')
    build_dir = os.path.join(doc_dir, 'build', 'html')

    args = ['-E',
            '-b',
            'html',
            '-d',
            doctrees_dir,
            source_dir,
            build_dir]
    status = main(args)

    import webbrowser
    webbrowser.open(os.path.join(build_dir, 'index.html'))




import unittest
import pydoc
from static_frame.test.test_case import TestCase

class TestUnit(TestCase):

    # this test is slow
    def test_interface_help_a(self) -> None:

        for target in self.get_containers():
            post = pydoc.render_doc(target, renderer=pydoc.plaintext) # type: ignore
            self.assertTrue(len(post) > 0)

if __name__ == '__main__':
    unittest.main()





# import typing as tp
import unittest
import numpy as np
from hypothesis import given

from static_frame.core.frame import Frame
from static_frame.core.index_datetime import IndexDate
from static_frame.core.archive_npy import NPYConverter
from static_frame.core.util import DTYPE_INEXACT_KINDS
from static_frame.test.property import strategies as sfst

from static_frame.test.test_case import temp_file
from static_frame.test.test_case import TestCase

class TestUnit(TestCase):

    @given(sfst.get_frame(dtype_group=sfst.DTGroup.ALL_NO_OBJECT,
            index_dtype_group=sfst.DTGroup.BASIC,
            columns_dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_npz_a(self, f1: Frame) -> None:
        # if f1.columns.dtype.kind != 'O' and f1.index.dtype.kind != 'O':
        with temp_file('.npz') as fp:
            f1.to_npz(fp)
            f2 = Frame.from_npz(fp)
            self.assertTrue(f1.equals(f2, compare_name=True, compare_dtype=True, compare_class=True))


    @given(sfst.get_frame(dtype_group=sfst.DTGroup.ALL_NO_OBJECT,
            index_cls=IndexDate,
            index_dtype_group=sfst.DTGroup.DATE,
            columns_cls=IndexDate,
            columns_dtype_group=sfst.DTGroup.DATE,
            ))
    def test_frame_to_npz_b(self, f1: Frame) -> None:
        # if f1.columns.dtype.kind != 'O' and f1.index.dtype.kind != 'O':
        with temp_file('.npz') as fp:
            f1.to_npz(fp)
            f2 = Frame.from_npz(fp)
            self.assertTrue(f1.equals(f2, compare_name=True, compare_dtype=True, compare_class=True))


    @given(sfst.get_array_1d2d(dtype_group=sfst.DTGroup.ALL_NO_OBJECT))
    def test_frame_to_npy_a(self, a1: Frame) -> None:
        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)

            # check compatibility with built-in NPY reading
            a2 = np.load(fp)
            if a2.dtype.kind in DTYPE_INEXACT_KINDS:
                self.assertAlmostEqualArray(a1, a2)
            else:
                self.assertTrue((a1 == a2).all())
            self.assertTrue(a1.shape == a2.shape)

            with open(fp, 'rb') as f:
                a3, _ = NPYConverter.from_npy(f)
                if a3.dtype.kind in DTYPE_INEXACT_KINDS:
                    self.assertAlmostEqualArray(a1, a3)
                else:
                    self.assertTrue((a1 == a3).all())
                self.assertTrue(a1.shape == a3.shape)

    @given(sfst.get_array_1d2d(dtype_group=sfst.DTGroup.ALL_NO_OBJECT))
    def test_frame_to_npy_b(self, a1: Frame) -> None:
        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)

            with open(fp, 'rb') as f:
                a2, mm = NPYConverter.from_npy(f, memory_map=True)
                if a2.dtype.kind in DTYPE_INEXACT_KINDS:
                    self.assertAlmostEqualArray(a1, a2)
                else:
                    self.assertTrue((a1 == a2).all())
                self.assertTrue(a1.shape == a2.shape)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import static_frame as sf
import frame_fixtures as ff


if __name__ == '__main__':

    # prlimit --as=850000000 python3 doc/source/articles/memmap_slice.py

    fp = '/tmp/big_frame'


    a1 = np.arange(10_000_000).reshape(1_000_000, 10)
    columns = tuple('abcdefghij')
    f1 = sf.Frame(a1, columns=columns)

    print('to npy')
    f1.to_npy(fp)

    # # loading two of these fails
    # print('start from_npy f2')
    # f2 = sf.Frame.from_npy(fp)

    # print('start from_npy f3')
    # f3 = sf.Frame.from_npy(fp)

    # we can create two of these

    print('start from_npy f2 from memmap')
    f2 = sf.Frame.from_npy(fp, memory_map=True)

    print('start from_npy f3 from memmap')
    f3 = sf.Frame.from_npy(fp, memory_map=True)

    for label, col in f3.items():
        print(label, col.shape)

    # print('start from_npy f4 from memmap')
    # f4 = sf.Frame.from_npy(fp, memory_map=True)





import zipfile
import os
import timeit
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import get_context as get_mp_context
import shutil
import typing as tp
from functools import partial

import numpy as np
import frame_fixtures as ff
import static_frame as sf
from static_frame.core.display_color import HexColor


# COUNT_ARRAY = 100
CHUNK_SIZE = 20


FF_100kx1k = 's(100_000,1_000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'
FF_10kx1k = 's(10_000,1_000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'
FF_100kx1k_auto = 's(100_000,1_000)|v(int,int,bool,float,float)'
FF_10kx1k_auto = 's(10_000,1_000)|v(int,int,bool,float,float)'

# FF_wide_col = 's(100,1_000)|v(int,bool,float)|i(I,int)|c(I,str)'
# FF_wide_ext = 's(1000,10_000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'

# FF_tall = 's(10_000,10)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'
# FF_tall_col = 's(10_000,10)|v(int,bool,float)|i(I,int)|c(I,str)'
# FF_tall_ext = 's(10_000,1000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'

# FF_square = 's(1_000,1_000)|v(float)|i(I,int)|c(I,str)'

def work(frame: sf.Frame) -> int:
    count = 0
    for row in frame.iter_series(axis=1):
        count += 1
    return count


class MMapTest:
    NUMBER = 1

    def __init__(self, fixture):
        self.fp_dir = '/tmp/npy'
        self.fixture = ff.parse(fixture)
        self.fixture.to_npy(self.fp_dir)

        partitions = 100
        ref = np.arange(len(self.fixture.columns))
        self.selections = [(ref % partitions) == i for i in range(partitions)]
        # import ipdb; ipdb.set_trace()
        print(self, 'init complete')

    def clear(self) -> None:
        if os.path.exists(self.fp_dir):
            shutil.rmtree(self.fp_dir)
        print(self, 'del complete')



# class CopyThread(MMapTest):

#     @staticmethod
#     def func(args):
#         frame, sel = args
#         return work(frame[sel])

#     def __call__(self):
#         with ThreadPoolExecutor() as executor:
#             args = ((self.fixture, sel) for sel in self.selections)
#             post = tuple(executor.map(self.func, args, chunksize=CHUNK_SIZE))
#             assert self.fixture.shape[0] == post[0]
#             assert len(post) == len(self.selections)

class CopyFork(MMapTest):

    @staticmethod
    def func(args):
        frame, sel = args
        return work(frame[sel])

    def __call__(self):
        with ProcessPoolExecutor(mp_context=get_mp_context('fork')) as executor:
            args = ((self.fixture, sel) for sel in self.selections)
            post = tuple(executor.map(self.func, args, chunksize=CHUNK_SIZE))
            assert self.fixture.shape[0] == post[0]
            assert len(post) == len(self.selections)

class CopySpawn(MMapTest):

    @staticmethod
    def func(args):
        frame, sel = args
        return work(frame[sel])

    def __call__(self):
        with ProcessPoolExecutor(mp_context=get_mp_context('spawn')) as executor:
            args = ((self.fixture, sel) for sel in self.selections)
            post = tuple(executor.map(self.func, args, chunksize=CHUNK_SIZE))
            assert self.fixture.shape[0] == post[0]
            assert len(post) == len(self.selections)



# class MMapThread(MMapTest):

#     @staticmethod
#     def func(fp: str, cols=np.ndarray):
#         return work(sf.Frame.from_npy(fp)[cols])

#     def __call__(self):
#         fp = self.fp_dir
#         func = partial(self.func, fp)

#         with ThreadPoolExecutor() as executor:
#             post = tuple(executor.map(func, self.selections, chunksize=CHUNK_SIZE))
#             assert self.fixture.shape[0] == post[0]
#             assert len(post) == len(self.selections)

class MMapFork(MMapTest):

    @staticmethod
    def func(fp: str, cols=np.ndarray):
        return work(sf.Frame.from_npy(fp)[cols])

    def __call__(self):
        fp = self.fp_dir
        func = partial(self.func, fp)

        with ProcessPoolExecutor(mp_context=get_mp_context('fork')) as executor:
            post = tuple(executor.map(func, self.selections, chunksize=CHUNK_SIZE))
            assert self.fixture.shape[0] == post[0]
            assert len(post) == len(self.selections)

class MMapSpawn(MMapTest):

    @staticmethod
    def func(fp: str, cols=np.ndarray):
        return work(sf.Frame.from_npy(fp)[cols])

    def __call__(self):
        fp = self.fp_dir
        func = partial(self.func, fp)

        with ProcessPoolExecutor(mp_context=get_mp_context('spawn')) as executor:
            post = tuple(executor.map(func, self.selections, chunksize=CHUNK_SIZE))
            assert self.fixture.shape[0] == post[0]
            assert len(post) == len(self.selections)





#-------------------------------------------------------------------------------
def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format


def run_test():
    records = []
    for label, fixture in (
            ('100kx1k', FF_100kx1k),
            ('10kx1k', FF_10kx1k),
            ('100kx1k_auto', FF_100kx1k_auto),
            ('10kx1k_auto', FF_10kx1k_auto),            ):
        for cls in (
                # CopyThread,
                CopyFork,
                # CopySpawn,
                # MMapThread,
                MMapFork,
                # MMapSpawn,
                ):
            runner = cls(fixture)
            record = [cls.__name__, cls.NUMBER, label]
            result = timeit.timeit(
                    f'runner()',
                    globals=locals(),
                    number=cls.NUMBER)
            record.append(result)
            records.append(record)
            runner.clear()

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'fixture', 'time')
            )

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))
    # import ipdb; ipdb.set_trace()



if __name__ == '__main__':
    run_test()




import typing as tp
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
# import multiprocessing as mp
# mp_context = mp.get_context('spawn')

import numpy as np

from static_frame.core.bus import Bus
from static_frame.core.container import ContainerOperand
from static_frame.core.display import Display
from static_frame.core.display import DisplayActive
from static_frame.core.display import DisplayHeader
from static_frame.core.display_config import DisplayConfig
from static_frame.core.doc_str import doc_inject
from static_frame.core.frame import Frame
from static_frame.core.index_auto import IndexAutoFactoryType
from static_frame.core.node_selector import InterfaceGetItem
from static_frame.core.node_selector import InterfaceSelectTrio
from static_frame.core.series import Series
from static_frame.core.store import Store
from static_frame.core.store import StoreConfigMap
from static_frame.core.store import StoreConfigMapInitializer
from static_frame.core.store_client_mixin import StoreClientMixin
from static_frame.core.store_hdf5 import StoreHDF5
from static_frame.core.store_sqlite import StoreSQLite
from static_frame.core.store_xlsx import StoreXLSX
from static_frame.core.store_zip import StoreZipCSV
from static_frame.core.store_zip import StoreZipParquet
from static_frame.core.store_zip import StoreZipPickle
from static_frame.core.store_zip import StoreZipTSV
from static_frame.core.store_zip import StoreZipNPZ
from static_frame.core.util import AnyCallable
from static_frame.core.util import Bloc2DKeyType
from static_frame.core.util import DEFAULT_SORT_KIND
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import ELEMENT_TUPLE
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import GetItemKeyTypeCompound
from static_frame.core.util import IndexInitializer
from static_frame.core.util import KeyOrKeys
from static_frame.core.util import NameType
from static_frame.core.util import PathSpecifier
from static_frame.core.util import UFunc
from static_frame.core.style_config import StyleConfig
from static_frame.core.exception import BatchIterableInvalid


FrameOrSeries = tp.Union[Frame, Series]
IteratorFrameItems = tp.Iterator[tp.Tuple[tp.Hashable, FrameOrSeries]]
GeneratorFrameItems = tp.Callable[..., IteratorFrameItems]


#-------------------------------------------------------------------------------
# family of executor functions normalized in signature (taking a single tuple of args) for usage in processor pool calls

def normalize_container(post: tp.Any
        ) -> FrameOrSeries:
    # post might be an element, promote to a Series to permit concatenation
    if post.__class__ is np.ndarray:
        if post.ndim == 1:
            return Series(post)
        elif post.ndim == 2:
            return Frame(post)
        # let ndim 0 pass
    if not isinstance(post, (Frame, Series)):
        # NOTE: do not set index as (container.name,), as this can lead to diagonal formations; will already be paired with stored labels
        return Series.from_element(post, index=ELEMENT_TUPLE)
    return post

def call_func(bundle: tp.Tuple[FrameOrSeries, AnyCallable]
        ) -> FrameOrSeries:
    container, func = bundle
    return normalize_container(func(container))

def call_func_items(bundle: tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]
        ) -> FrameOrSeries:
    container, func, label = bundle
    return normalize_container(func(label, container))

def call_attr(bundle: tp.Tuple[FrameOrSeries, str, tp.Any, tp.Any]
        ) -> FrameOrSeries:
    container, attr, args, kwargs = bundle
    func = getattr(container, attr)
    return normalize_container(func(*args, **kwargs))

#-------------------------------------------------------------------------------
class Batch(ContainerOperand, StoreClientMixin):
    '''
    A lazy, sequentially evaluated container of :obj:`Frame` that broadcasts operations on contained :obj:`Frame` by return new :obj:`Batch` instances. Full evaluation of operations only occurs when iterating or calling an exporter.
    '''

    __slots__ = (
            '_items',
            '_name',
            '_config',
            '_max_workers',
            '_chunksize',
            '_use_threads',
            )

    _config: StoreConfigMap

    @classmethod
    def from_frames(cls,
            frames: tp.Iterable[Frame],
            *,
            name: NameType = None,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''Return a :obj:`Batch` from an iterable of :obj:`Frame`; labels will be drawn from :obj:`Frame.name`.
        '''
        return cls(((f.name, f) for f in frames),
                name=name,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    #---------------------------------------------------------------------------
    # constructors by data format

    @classmethod
    def _from_store(cls,
            store: Store,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        config_map = StoreConfigMap.from_initializer(config)

        items = ((label, store.read(label, config=config_map[label]))
                for label in store.labels(config=config_map))

        return cls(items,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_tsv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped TSV :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipTSV(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_csv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped CSV :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipCSV(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_pickle(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped pickle :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipPickle(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_npz(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped NPZ :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipNPZ(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_parquet(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped parquet :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipParquet(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_xlsx(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to an XLSX :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        # how to pass configuration for multiple sheets?
        store = StoreXLSX(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_sqlite(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to an SQLite :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreSQLite(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_hdf5(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to a HDF5 :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreHDF5(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    #---------------------------------------------------------------------------
    @doc_inject(selector='batch_init')
    def __init__(self,
            items: IteratorFrameItems,
            *,
            name: NameType = None,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ):
        '''
        Default constructor of a :obj:`Batch`.

        {args}
        '''
        self._items = items # might be a generator!
        self._name = name

        self._config = StoreConfigMap.from_initializer(config)

        self._max_workers = max_workers
        self._chunksize = chunksize
        self._use_threads = use_threads

    #---------------------------------------------------------------------------
    def _derive(self,
            gen: GeneratorFrameItems,
            name: NameType = None,
            ) -> 'Batch':
        '''Utility for creating derived Batch
        '''
        return self.__class__(gen(),
                name=name if name is not None else self._name,
                config=self._config,
                max_workers=self._max_workers,
                chunksize=self._chunksize,
                use_threads=self._use_threads,
                )

    #---------------------------------------------------------------------------
    # name interface

    @property #type: ignore
    @doc_inject()
    def name(self) -> NameType:
        '''{}'''
        return self._name

    def rename(self, name: NameType) -> 'Batch':
        '''
        Return a new Batch with an updated name attribute.
        '''
        def gen() -> IteratorFrameItems:
            yield from self._items
        return self._derive(gen, name=name)

    #---------------------------------------------------------------------------
    @property
    def shapes(self) -> Series:
        '''A :obj:`Series` describing the shape of each iterated :obj:`Frame`.

        Returns:
            :obj:`tp.Tuple[int]`
        '''
        items = ((label, f.shape) for label, f in self._items)
        return Series.from_items(items, name='shape', dtype=DTYPE_OBJECT)


    def display(self,
            config: tp.Optional[DisplayConfig] = None,
            *,
            style_config: tp.Optional[StyleConfig] = None,
            ) -> Display:
        '''Provide a :obj:`Series`-style display of the :obj:`Batch`. Note that if the held iterator is a generator, this display will exhaust the generator.
        '''
        config = config or DisplayActive.get()

        items = ((label, f.__class__) for label, f in self._items)
        series = Series.from_items(items, name=self._name)

        display_cls = Display.from_values((),
                header=DisplayHeader(self.__class__, self._name),
                config=config)
        return series._display(config,
                display_cls=display_cls,
                style_config=style_config,
                )

    def __repr__(self) -> str:
        '''Provide a display of the :obj:`Batch` that does not exhaust the generator.
        '''
        if self._name:
            header = f'{self.__class__.__name__}: {self._name}'
        else:
            header = self.__class__.__name__
        return f'<{header} at {hex(id(self))}>'

    #---------------------------------------------------------------------------
    # core function application routines

    def _iter_items(self) -> IteratorFrameItems:
        '''Iter pairs in items, providing helpful exception of a pair is not found. Thies is necessary as we cannot validate the items until we actually do an iteration, and the iterable might be an iterator.
        '''
        for pair in self._items:
            try:
                label, frame = pair
            except ValueError:
                raise BatchIterableInvalid() from None
            yield label, frame

    def _apply_pool(self,
            labels: tp.List[tp.Hashable],
            arg_iter: tp.Iterator[tp.Tuple[tp.Any, ...]],
            caller: tp.Callable[..., FrameOrSeries],
            ) -> 'Batch':

        pool_executor = ThreadPoolExecutor if self._use_threads else ProcessPoolExecutor

        def gen_pool() -> IteratorFrameItems:
            with pool_executor(max_workers=self._max_workers) as executor:
                yield from zip(labels,
                        executor.map(caller, arg_iter, chunksize=self._chunksize)
                        )
        return self._derive(gen_pool)

    def _apply_pool_except(self,
            labels: tp.List[tp.Hashable],
            arg_iter: tp.Iterator[tp.Tuple[tp.Any, ...]],
            caller: tp.Callable[..., FrameOrSeries],
            exception: tp.Type[Exception],
            ) -> 'Batch':

        if self._chunksize != 1:
            raise NotImplementedError('Cannot use apply_except idioms with chunksize other than 1')

        pool_executor = ThreadPoolExecutor if self._use_threads else ProcessPoolExecutor

        def gen_pool() -> IteratorFrameItems:
            futures = []
            with pool_executor(max_workers=self._max_workers) as executor:
                for args in arg_iter:
                    futures.append(executor.submit(caller, args))

                for label, future in zip(labels, futures):
                    try:
                        container = future.result()
                    except exception:
                        continue
                    yield label, container

        return self._derive(gen_pool)

    def _apply_attr(self,
            *args: tp.Any,
            attr: str,
            **kwargs: tp.Any,
            ) -> 'Batch':
        '''
        Apply a method on a Frame given as an attr string.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._iter_items():
                    yield label, call_attr((frame, attr, args, kwargs))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, str, tp.Any, tp.Any]]:
            for label, frame in self._iter_items():
                labels.append(label)
                yield frame, attr, args, kwargs

        return self._apply_pool(labels, arg_gen(), call_attr)

    def apply(self, func: AnyCallable) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the :obj:`Frame` as an argument.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._iter_items():
                    yield label, call_func((frame, func))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable]]:
            for label, frame in self._iter_items():
                labels.append(label)
                yield frame, func

        return self._apply_pool(labels, arg_gen(), call_func)

    def apply_except(self,
            func: AnyCallable,
            exception: tp.Type[Exception],
            ) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the :obj:`Frame` as an argument. Exceptions raised that matching the `except` argument will be silenced.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._iter_items():
                    try:
                        yield label, call_func((frame, func))
                    except exception:
                        pass
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable]]:
            for label, frame in self._iter_items():
                labels.append(label)
                yield frame, func

        return self._apply_pool_except(labels,
                arg_gen(),
                call_func,
                exception,
                )

    def apply_items(self, func: AnyCallable) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the pair of label, :obj:`Frame` as an argument.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._iter_items():
                    yield label, call_func_items((frame, func, label))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]]:
            for label, frame in self._iter_items():
                labels.append(label)
                yield frame, func, label

        return self._apply_pool(labels, arg_gen(), call_func_items)

    def apply_items_except(self,
            func: AnyCallable,
            exception: tp.Type[Exception],
            ) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the pair of label, :obj:`Frame` as an argument. Exceptions raised that matching the `except` argument will be silenced.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._iter_items():
                    try:
                        yield label, call_func_items((frame, func, label))
                    except exception:
                        pass
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]]:
            for label, frame in self._iter_items():
                labels.append(label)
                yield frame, func, label

        return self._apply_pool_except(labels,
                arg_gen(),
                call_func_items,
                exception,
                )

    #---------------------------------------------------------------------------
    # extraction

    def _extract_iloc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_extract_iloc',
                key=key
                )
    def _extract_loc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_extract_loc',
                key=key
                )

    def _extract_bloc(self, key: Bloc2DKeyType) -> 'Batch':
        return self._apply_attr(
                attr='_extract_bloc',
                key=key
                )

    def __getitem__(self, key: GetItemKeyType) -> 'Batch':
        ''
        return self._apply_attr(
                attr='__getitem__',
                key=key
                )

    #---------------------------------------------------------------------------
    def _drop_iloc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_iloc',
                key=key
                )

    def _drop_loc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_loc',
                key=key
                )

    def _drop_getitem(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_getitem',
                key=key
                )

    #---------------------------------------------------------------------------
    # interfaces

    @property
    def loc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_loc)

    @property
    def iloc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_iloc)

    @property
    def bloc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_bloc)

    @property
    def drop(self) -> InterfaceSelectTrio['Batch']:
        return InterfaceSelectTrio( #type: ignore
            func_iloc=self._drop_iloc,
            func_loc=self._drop_loc,
            func_getitem=self._drop_getitem)

    # NOTE: note sure if assign interfaces would work in this context

    #---------------------------------------------------------------------------
    # dictionary-like interface
    # these methods operate on the Batch itself, not the contained Frames

    def keys(self) -> tp.Iterator[tp.Hashable]:
        '''
        Iterator of :obj:`Frame` labels.
        '''
        for k, _ in self._iter_items():
            yield k

    def __iter__(self) -> tp.Iterator[tp.Hashable]:
        '''
        Iterator of :obj:`Frame` labels, same as :obj:`Batch.keys`.
        '''
        yield from self.keys()

    @property
    def values(self) -> tp.Iterator[FrameOrSeries]:
        '''
        Return an iterator of values (:obj:`Frame` or :obj:`Series`) stored in this :obj:`Batch`.
        '''
        return (v for _, v in self._iter_items())

    def items(self) -> IteratorFrameItems:
        '''
        Iterator of labels, :obj:`Frame`.
        '''
        return self._iter_items()

    _items_store = items

    #---------------------------------------------------------------------------
    # axis and shape ufunc methods

    def _ufunc_unary_operator(self,
            operator: UFunc
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_unary_operator',
                operator=operator
                )

    def _ufunc_binary_operator(self, *,
            operator: UFunc,
            other: tp.Any,
            fill_value: object = np.nan,
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_binary_operator',
                operator=operator,
                other=other,
                )

    def _ufunc_axis_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_axis_skipna',
                axis=axis,
                skipna=skipna,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna,
                composable=composable,
                dtypes=dtypes,
                size_one_unity=size_one_unity,
                )

    def _ufunc_shape_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> 'Batch':

        return self._apply_attr(
                attr='_ufunc_shape_skipna',
                axis=axis,
                skipna=skipna,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna,
                composable=composable,
                dtypes=dtypes,
                size_one_unity=size_one_unity,
                )

    #---------------------------------------------------------------------------
    # transformations resulting in the same dimensionality

    def sort_index(self,
            *,
            ascending: bool = True,
            kind: str = DEFAULT_SORT_KIND
            ) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj;`Frame` ordered by the sorted ``index``.
        '''
        return self._apply_attr(
                attr='sort_index',
                ascending=ascending,
                kind=kind,
                )

    def sort_columns(self,
            *,
            ascending: bool = True,
            kind: str = DEFAULT_SORT_KIND
            ) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` ordered by the sorted ``columns``.
        '''
        return self._apply_attr(
                attr='sort_columns',
                ascending=ascending,
                kind=kind,
                )

    def sort_values(self,
            label: KeyOrKeys,
            *,
            ascending: bool = True,
            axis: int = 1,
            kind: str = DEFAULT_SORT_KIND) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` ordered by the sorted values, where values are given by single column or iterable of columns.

        Args:
            label: a label or iterable of keys.
        '''
        return self._apply_attr(
                attr='sort_values',
                label=label,
                ascending=ascending,
                axis=axis,
                kind=kind,
                )

    def isin(self, other: tp.Any) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` as a same-sized Boolean :obj:`Frame` that shows if the same-positioned element is in the passed iterable.
        '''
        return self._apply_attr(
                attr='isin',
                other=other,
                )

    @doc_inject(class_name='Batch')
    def clip(self, *,
            lower: tp.Optional[tp.Union[float, Series, Frame]] = None,
            upper: tp.Optional[tp.Union[float, Series, Frame]] = None,
            axis: tp.Optional[int] = None
            ) -> 'Batch':
        '''{}

        Args:
            lower: value, :obj:`Series`, :obj:`Frame`
            upper: value, :obj:`Series`, :obj:`Frame`
            axis: required if ``lower`` or ``upper`` are given as a :obj:`Series`.
        '''
        return self._apply_attr(
                attr='clip',
                lower=lower,
                upper=upper,
                axis=axis,
                )

    def transpose(self) -> 'Batch':
        '''Transpose. Return a :obj:`Frame` with ``index`` as ``columns`` and vice versa.
        '''
        return self._apply_attr(attr='transpose')

    @property
    def T(self) -> 'Batch':
        '''Transpose. Return a :obj:`Frame` with ``index`` as ``columns`` and vice versa.
        '''
        return self._apply_attr(attr='transpose')



    @doc_inject(selector='duplicated')
    def duplicated(self, *,
            axis: int = 0,
            exclude_first: bool = False,
            exclude_last: bool = False) -> 'Batch':
        '''
        Return an axis-sized Boolean :obj:`Series` that shows True for all rows (axis 0) or columns (axis 1) duplicated.

        Args:
            {axis}
            {exclude_first}
            {exclude_last}
        '''
        return self._apply_attr(
                attr='duplicated',
                axis=axis,
                exclude_first=exclude_first,
                exclude_last=exclude_last,
                )

    @doc_inject(selector='duplicated')
    def drop_duplicated(self, *,
            axis: int = 0,
            exclude_first: bool = False,
            exclude_last: bool = False
            ) -> 'Batch':
        '''
        Return a :obj:`Batch` with contained :obj:`Frame` with duplicated rows (axis 0) or columns (axis 1) removed. All values in the row or column are compared to determine duplication.

        Args:
            {axis}
            {exclude_first}
            {exclude_last}
        '''
        return self._apply_attr(
                attr='drop_duplicated',
                axis=axis,
                exclude_first=exclude_first,
                exclude_last=exclude_last,
                )

    # as only useful on Frame, perhaps skip?
    # def set_index(self,
    # def set_index_hierarchy(self,
    # def unset_index(self, *,

    def __round__(self, decimals: int = 0) -> 'Batch':
        '''
        Return a :obj:`Batch` with contained :obj:`Frame` rounded to the given decimals. Negative decimals round to the left of the decimal point.

        Args:
            decimals: number of decimals to round to.
        '''
        return self._apply_attr(
                attr='__round__',
                decimals=decimals,
                )

    def roll(self,
            index: int = 0,
            columns: int = 0,
            include_index: bool = False,
            include_columns: bool = False,
            ) -> 'Batch':
        '''
        Roll columns and/or rows by positive or negative integer counts, where columns and/or rows roll around the axis.

        Args:
            include_index: Determine if index is included in index-wise rotation.
            include_columns: Determine if column index is included in index-wise rotation.
        '''
        return self._apply_attr(
                attr='roll',
                index=index,
                columns=columns,
                include_index=include_index,
                include_columns=include_columns,
                )

    def shift(self,
            index: int = 0,
            columns: int = 0,
            fill_value: tp.Any = np.nan) -> 'Batch':
        '''
        Shift columns and/or rows by positive or negative integer counts, where columns and/or rows fall of the axis and introduce missing values, filled by `fill_value`.
        '''
        return self._apply_attr(
                attr='shift',
                index=index,
                columns=columns,
                fill_value=fill_value,
                )

    #---------------------------------------------------------------------------
    # transformations resulting in changed dimensionality

    def count(self, *,
            skipna: bool = True,
            axis: int = 0,
            ) -> 'Batch':
        '''Apply count on contained Frames.
        '''
        return self._apply_attr(
                attr='count',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='sample')
    def sample(self,
            index: tp.Optional[int] = None,
            columns: tp.Optional[int] = None,
            *,
            seed: tp.Optional[int] = None,
            ) -> 'Batch':
        '''Apply sample on contained Frames.

        Args:
            {index}
            {columns}
            {seed}
        '''
        return self._apply_attr(
                attr='sample',
                index=index,
                columns=columns,
                seed=seed,
                )

    @doc_inject(selector='head', class_name='Batch')
    def head(self, count: int = 5) -> 'Batch':
        '''{doc}

        Args:
            {count}
        '''
        return self._apply_attr(
                attr='head',
                count=count,
                )

    @doc_inject(selector='tail', class_name='Batch')
    def tail(self, count: int = 5) -> 'Batch':
        '''{doc}

        Args:
            {count}
        '''
        return self._apply_attr(
                attr='tail',
                count=count,
                )

    @doc_inject(selector='argminmax')
    def loc_min(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the labels corresponding to the minimum value found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='loc_min',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def iloc_min(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the integer indices corresponding to the minimum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='iloc_min',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def loc_max(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the labels corresponding to the maximum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='loc_max',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def iloc_max(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the integer indices corresponding to the maximum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='iloc_max',
                skipna=skipna,
                axis=axis,
                )

    def cov(self, *,
            axis: int = 1,
            ddof: int = 1,
            ) -> 'Batch':
        '''
        Compute a covariance matrix.

        Args:
            axis: if 0, each row represents a variable, with observations as columns; if 1, each column represents a variable, with observations as rows. Defaults to 1.
            ddof: Delta degrees of freedom, defaults to 1.
        '''
        return self._apply_attr(
                attr='cov',
                axis=axis,
                ddof=ddof,
                )


    #---------------------------------------------------------------------------
    # utility function to numpy array

    def unique(self, *,
            axis: tp.Optional[int] = None,
            ) -> 'Batch':
        '''
        Return a NumPy array of unqiue values. If the axis argument is provied, uniqueness is determined by columns or row.

        '''
        return self._apply_attr(
                attr='unique',
                axis=axis,
                )

    #---------------------------------------------------------------------------
    # exporter

    def to_frame(self, *,
            axis: int = 0,
            union: bool = True,
            index: tp.Optional[tp.Union[IndexInitializer, IndexAutoFactoryType]] = None,
            columns: tp.Optional[tp.Union[IndexInitializer, IndexAutoFactoryType]] = None,
            name: NameType = None,
            fill_value: object = np.nan,
            consolidate_blocks: bool = False
        ) -> Frame:
        '''
        Consolidate stored :obj:`Frame` into a new :obj:`Frame` using the stored labels as the index on the provided ``axis`` using :obj:`Frame.from_concat`. This assumes that that the contained :obj:`Frame` have been reduced to single dimension along the provided `axis`.
        '''
        labels = []
        containers: tp.List[FrameOrSeries] = []
        ndim1d = True
        for label, container in self._items:
            labels.append(label)
            ndim1d &= container.ndim == 1
            containers.append(container)

        name = name if name is not None else self._name

        if ndim1d:
            if axis == 0 and index is None:
                index = labels
            if axis == 1 and columns is None:
                columns = labels
            # import ipdb; ipdb.set_trace()
            return Frame.from_concat( #type: ignore
                    containers,
                    axis=axis,
                    union=union,
                    index=index,
                    columns=columns,
                    name=name,
                    fill_value=fill_value,
                    consolidate_blocks=consolidate_blocks,
                    )
        # produce a hierarchical index to return all Frames
        f = Frame.from_concat_items(
                zip(labels, containers),
                axis=axis,
                union=union,
                name=name,
                fill_value=fill_value,
                consolidate_blocks=consolidate_blocks,
                )
        if index is not None or columns is not None:
            # this relabels, as that is how Frame.from_concat works
            f = f.relabel(index=index, columns=columns)
        return f

    def to_bus(self) -> 'Bus':
        '''Realize the :obj:`Batch` as an :obj:`Bus`. Note that, as a :obj:`Bus` must have all labels (even if :obj:`Frame` are loaded lazily), this :obj:`Batch` will be exhausted.
        '''
        series = Series.from_items(
                self.items(),
                name=self._name,
                dtype=DTYPE_OBJECT)

        return Bus(series, config=self._config)







from static_frame import Frame
from static_frame.test.test_case import TestCase
from static_frame.test.test_case import skip_linux_no_display


class TestUnit(TestCase):

    # NOTE: this test will end up clearing the user's clipboard

    @skip_linux_no_display #type: ignore
    def test_frame_to_clipboard_a(self) -> None:
        records = (
                (2, 'a', False),
                (3, 'b', False),
                )
        f1 = Frame.from_records(records,
                columns=('r', 's', 't'),
                index=('w', 'x'))

        f1.to_clipboard()
        f2 = Frame.from_clipboard(index_depth=1)
        self.assertTrue(f2.equals(f1, compare_dtype=True))



